import sys
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from typing import AbstractSet, Any, Callable, NamedTuple, Optional, Union, cast  # noqa: UP035

import dagster._check as check
from dagster._core.definitions import IJob, JobDefinition
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.job_base import InMemoryJob
from dagster._core.definitions.reconstruct import ReconstructableJob
from dagster._core.definitions.repository_definition import RepositoryLoadData
from dagster._core.errors import DagsterExecutionInterruptedError, DagsterInvariantViolationError
from dagster._core.events import DagsterEvent, EngineEventData, JobFailureData, RunFailureReason
from dagster._core.execution.context.system import PlanOrchestrationContext
from dagster._core.execution.context_creation_job import (
    ExecutionContextManager,
    PlanExecutionContextManager,
    PlanOrchestrationContextManager,
    orchestration_context_event_generator,
    scoped_job_context,
)
from dagster._core.execution.job_execution_result import JobExecutionResult
from dagster._core.execution.plan.execute_plan import inner_plan_execution_iterator
from dagster._core.execution.plan.plan import ExecutionPlan
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.execution.retries import RetryMode
from dagster._core.instance import DagsterInstance, InstanceRef
from dagster._core.selector import parse_step_selection
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus
from dagster._core.system_config.objects import ResolvedRunConfig
from dagster._core.telemetry import log_dagster_event, log_repo_stats, telemetry_wrapper
from dagster._utils.error import serializable_error_info_from_exc_info
from dagster._utils.interrupts import capture_interrupts
from dagster._utils.merger import merge_dicts

## Brief guide to the execution APIs
# | function name               | operates over      | sync  | supports    | creates new DagsterRun  |
# |                             |                    |       | reexecution | in instance             |
# | --------------------------- | ------------------ | ----- | ----------- | ----------------------- |
# | execute_job                 | ReconstructableJob | sync  | yes         | yes                     |
# | execute_run_iterator        | DagsterRun         | async | (1)         | no                      |
# | execute_run                 | DagsterRun         | sync  | (1)         | no                      |
# | execute_plan_iterator       | ExecutionPlan      | async | (2)         | no                      |
# | execute_plan                | ExecutionPlan      | sync  | (2)         | no                      |
#
# Notes on reexecution support:
# (1) The appropriate bits must be set on the DagsterRun passed to this function. Specifically,
#     parent_run_id and root_run_id must be set and consistent, and if a resolved_op_selection or
#     step_keys_to_execute are set they must be consistent with the parent and root runs.
# (2) As for (1), but the ExecutionPlan passed must also agree in all relevant bits.


def execute_run_iterator(
    job: IJob,
    dagster_run: DagsterRun,
    instance: DagsterInstance,
    resume_from_failure: bool = False,
) -> Iterator[DagsterEvent]:
    check.inst_param(job, "job", IJob)
    check.inst_param(dagster_run, "dagster_run", DagsterRun)
    check.inst_param(instance, "instance", DagsterInstance)

    if dagster_run.status == DagsterRunStatus.CANCELED:
        # This can happen if the run was force-terminated while it was starting
        def gen_execute_on_cancel():
            yield instance.report_engine_event(
                "Not starting execution since the run was canceled before execution could start",
                dagster_run,
            )

        return gen_execute_on_cancel()

    if not resume_from_failure:
        if dagster_run.status not in (DagsterRunStatus.NOT_STARTED, DagsterRunStatus.STARTING):
            if dagster_run.is_finished:

                def gen_ignore_duplicate_run_worker():
                    yield instance.report_engine_event(
                        "Ignoring a run worker that started after the run had already finished.",
                        dagster_run,
                    )

                return gen_ignore_duplicate_run_worker()
            elif instance.run_monitoring_enabled:
                # This can happen if the pod was unexpectedly restarted by the cluster - ignore it since
                # the run monitoring daemon will also spin up a new pod
                def gen_ignore_duplicate_run_worker():
                    yield instance.report_engine_event(
                        "Ignoring a duplicate run that was started from somewhere other than"
                        " the run monitor daemon",
                        dagster_run,
                    )

                return gen_ignore_duplicate_run_worker()
            else:

                def gen_fail_restarted_run_worker():
                    yield instance.report_engine_event(
                        f"{dagster_run.job_name} ({dagster_run.run_id}) started a new"
                        f" run worker while the run was already in state {dagster_run.status}."
                        " This most frequently happens when the run worker unexpectedly stops"
                        " and is restarted by the cluster. Marking the run as failed.",
                        dagster_run,
                    )
                    yield instance.report_run_failed(
                        dagster_run,
                        job_failure_data=JobFailureData(
                            error=None, failure_reason=RunFailureReason.RUN_WORKER_RESTART
                        ),
                    )

                return gen_fail_restarted_run_worker()

    else:
        check.invariant(
            dagster_run.status == DagsterRunStatus.STARTED
            or dagster_run.status == DagsterRunStatus.STARTING,
            desc=(
                f"Run of {dagster_run.job_name} ({dagster_run.run_id}) in state {dagster_run.status}, expected STARTED or STARTING because it's "
                "resuming from a run worker failure"
            ),
        )

    if (
        dagster_run.resolved_op_selection
        or dagster_run.asset_selection
        or dagster_run.asset_check_selection
    ):
        # when `execute_run_iterator` is directly called, the sub pipeline hasn't been created
        # note that when we receive the solids to execute via DagsterRun, it won't support
        # solid selection query syntax
        job = job.get_subset(
            op_selection=(
                list(dagster_run.resolved_op_selection)
                if dagster_run.resolved_op_selection
                else None
            ),
            asset_selection=dagster_run.asset_selection,
            asset_check_selection=dagster_run.asset_check_selection,
        )

    execution_plan = _get_execution_plan_from_run(job, dagster_run, instance)
    if isinstance(job, ReconstructableJob):
        job = job.with_repository_load_data(execution_plan.repository_load_data)

    return iter(
        ExecuteRunWithPlanIterable(
            execution_plan=execution_plan,
            iterator=job_execution_iterator,
            execution_context_manager=PlanOrchestrationContextManager(
                context_event_generator=orchestration_context_event_generator,
                job=job,
                execution_plan=execution_plan,
                dagster_run=dagster_run,
                instance=instance,
                run_config=dagster_run.run_config,
                raise_on_error=False,
                executor_defs=None,
                output_capture=None,
                resume_from_failure=resume_from_failure,
            ),
        )
    )


def execute_run(
    job: IJob,
    dagster_run: DagsterRun,
    instance: DagsterInstance,
    raise_on_error: bool = False,
) -> JobExecutionResult:
    """Executes an existing job run synchronously.

    Synchronous version of execute_run_iterator.

    Args:
        job (IJob): The pipeline to execute.
        dagster_run (DagsterRun): The run to execute
        instance (DagsterInstance): The instance in which the run has been created.
        raise_on_error (Optional[bool]): Whether or not to raise exceptions when they occur.
            Defaults to ``False``.

    Returns:
        JobExecutionResult: The result of the execution.
    """
    if isinstance(job, JobDefinition):
        raise DagsterInvariantViolationError(
            "execute_run requires a reconstructable job but received job definition directly"
            " instead. To support hand-off to other processes please wrap your definition in a call"
            " to reconstructable(). Learn more about reconstructable here:"
            " https://docs.dagster.io/api/python-api/execution#dagster.reconstructable"
        )

    check.inst_param(job, "job", IJob)
    check.inst_param(dagster_run, "dagster_run", DagsterRun)
    check.inst_param(instance, "instance", DagsterInstance)

    if dagster_run.status == DagsterRunStatus.CANCELED:
        message = "Not starting execution since the run was canceled before execution could start"
        instance.report_engine_event(
            message,
            dagster_run,
        )
        raise DagsterInvariantViolationError(message)

    check.invariant(
        dagster_run.status == DagsterRunStatus.NOT_STARTED
        or dagster_run.status == DagsterRunStatus.STARTING,
        desc=f"Run {dagster_run.job_name} ({dagster_run.run_id}) in state {dagster_run.status}, expected NOT_STARTED or STARTING",
    )
    if (
        dagster_run.resolved_op_selection
        or dagster_run.asset_selection
        or dagster_run.asset_check_selection
    ):
        # when `execute_run` is directly called, the sub job hasn't been created
        # note that when we receive the solids to execute via DagsterRun, it won't support
        # solid selection query syntax
        job = job.get_subset(
            op_selection=(
                list(dagster_run.resolved_op_selection)
                if dagster_run.resolved_op_selection
                else None
            ),
            asset_selection=dagster_run.asset_selection,
            asset_check_selection=dagster_run.asset_check_selection,
        )

    execution_plan = _get_execution_plan_from_run(job, dagster_run, instance)
    if isinstance(job, ReconstructableJob):
        job = job.with_repository_load_data(execution_plan.repository_load_data)

    _execute_run_iterable = ExecuteRunWithPlanIterable(
        execution_plan=execution_plan,
        iterator=job_execution_iterator,
        execution_context_manager=PlanOrchestrationContextManager(
            context_event_generator=orchestration_context_event_generator,
            job=job,
            execution_plan=execution_plan,
            dagster_run=dagster_run,
            instance=instance,
            run_config=dagster_run.run_config,
            raise_on_error=raise_on_error,
            executor_defs=None,
            output_capture=None,
        ),
    )
    event_list = list(_execute_run_iterable)

    # We need to reload the run object after execution for it to be accurate
    reloaded_dagster_run = check.not_none(instance.get_run_by_id(dagster_run.run_id))

    return JobExecutionResult(
        job.get_definition(),
        scoped_job_context(
            execution_plan,
            job,
            reloaded_dagster_run.run_config,
            reloaded_dagster_run,
            instance,
        ),
        event_list,
        reloaded_dagster_run,
    )


@contextmanager
def ephemeral_instance_if_missing(
    instance: Optional[DagsterInstance],
) -> Iterator[DagsterInstance]:
    if instance:
        yield instance
    else:
        with DagsterInstance.ephemeral() as ephemeral_instance:
            yield ephemeral_instance


class ReexecutionOptions(NamedTuple):
    """Reexecution options for python-based execution in Dagster.

    Args:
        parent_run_id (str): The run_id of the run to reexecute.
        step_selection (Sequence[str]):
            The list of step selections to reexecute. Must be a subset or match of the
            set of steps executed in the original run. For example:

            - ``['some_op']``: selects ``some_op`` itself.
            - ``['*some_op']``: select ``some_op`` and all its ancestors (upstream dependencies).
            - ``['*some_op+++']``: select ``some_op``, all its ancestors, and its descendants
              (downstream dependencies) within 3 levels down.
            - ``['*some_op', 'other_op_a', 'other_op_b+']``: select ``some_op`` and all its
              ancestors, ``other_op_a`` itself, and ``other_op_b`` and its direct child ops.
    """

    parent_run_id: str
    step_selection: Sequence[str] = []

    @staticmethod
    def from_failure(run_id: str, instance: DagsterInstance) -> "ReexecutionOptions":
        """Creates reexecution options from a failed run.

        Args:
            run_id (str): The run_id of the failed run. Run must fail in order to be reexecuted.
            instance (DagsterInstance): The DagsterInstance that the original run occurred in.

        Returns:
            ReexecutionOptions: Reexecution options to pass to a python execution.
        """
        return ReexecuteFromFailureOption(parent_run_id=run_id)


class ReexecuteFromFailureOption(ReexecutionOptions):
    """Marker subclass used to calculate reexecution information later."""


def execute_job(
    job: ReconstructableJob,
    instance: "DagsterInstance",
    run_config: Any = None,
    tags: Optional[Mapping[str, Any]] = None,
    raise_on_error: bool = False,
    op_selection: Optional[Sequence[str]] = None,
    reexecution_options: Optional[ReexecutionOptions] = None,
    asset_selection: Optional[Sequence[AssetKey]] = None,
) -> JobExecutionResult:
    """Execute a job synchronously.

    This API represents dagster's python entrypoint for out-of-process
    execution. For most testing purposes, :py:meth:`~dagster.JobDefinition.
    execute_in_process` will be more suitable, but when wanting to run
    execution using an out-of-process executor (such as :py:class:`dagster.
    multiprocess_executor`), then `execute_job` is suitable.

    `execute_job` expects a persistent :py:class:`DagsterInstance` for
    execution, meaning the `$DAGSTER_HOME` environment variable must be set.
    It also expects a reconstructable pointer to a :py:class:`JobDefinition` so
    that it can be reconstructed in separate processes. This can be done by
    wrapping the ``JobDefinition`` in a call to :py:func:`dagster.
    reconstructable`.

    .. code-block:: python

        from dagster import DagsterInstance, execute_job, job, reconstructable

        @job
        def the_job():
            ...

        instance = DagsterInstance.get()
        result = execute_job(reconstructable(the_job), instance=instance)
        assert result.success


    If using the :py:meth:`~dagster.GraphDefinition.to_job` method to
    construct the ``JobDefinition``, then the invocation must be wrapped in a
    module-scope function, which can be passed to ``reconstructable``.

    .. code-block:: python

        from dagster import graph, reconstructable

        @graph
        def the_graph():
            ...

        def define_job():
            return the_graph.to_job(...)

        result = execute_job(reconstructable(define_job), ...)

    Since `execute_job` is potentially executing outside of the current
    process, output objects need to be retrieved by use of the provided job's
    io managers. Output objects can be retrieved by opening the result of
    `execute_job` as a context manager.

    .. code-block:: python

        from dagster import execute_job

        with execute_job(...) as result:
            output_obj = result.output_for_node("some_op")

    ``execute_job`` can also be used to reexecute a run, by providing a :py:class:`ReexecutionOptions` object.

    .. code-block:: python

        from dagster import ReexecutionOptions, execute_job

        instance = DagsterInstance.get()

        options = ReexecutionOptions.from_failure(run_id=failed_run_id, instance)
        execute_job(reconstructable(job), instance, reexecution_options=options)

    Parameters:
        job (ReconstructableJob): A reconstructable pointer to a :py:class:`JobDefinition`.
        instance (DagsterInstance): The instance to execute against.
        run_config (Optional[dict]): The configuration that parametrizes this run, as a dict.
        tags (Optional[Dict[str, Any]]): Arbitrary key-value pairs that will be added to run logs.
        raise_on_error (Optional[bool]): Whether or not to raise exceptions when they occur.
            Defaults to ``False``.
        op_selection (Optional[List[str]]): A list of op selection queries (including single
            op names) to execute. For example:

            - ``['some_op']``: selects ``some_op`` itself.
            - ``['*some_op']``: select ``some_op`` and all its ancestors (upstream dependencies).
            - ``['*some_op+++']``: select ``some_op``, all its ancestors, and its descendants
              (downstream dependencies) within 3 levels down.
            - ``['*some_op', 'other_op_a', 'other_op_b+']``: select ``some_op`` and all its
              ancestors, ``other_op_a`` itself, and ``other_op_b`` and its direct child ops.
        reexecution_options (Optional[ReexecutionOptions]):
            Reexecution options to provide to the run, if this run is
            intended to be a reexecution of a previous run. Cannot be used in
            tandem with the ``op_selection`` argument.

    Returns:
      :py:class:`JobExecutionResult`: The result of job execution.
    """
    check.inst_param(job, "job", ReconstructableJob)
    check.inst_param(instance, "instance", DagsterInstance)
    check.opt_sequence_param(asset_selection, "asset_selection", of_type=AssetKey)

    # get the repository load data here because we call job.get_definition() later in this fn
    job_def, _ = _job_with_repository_load_data(job)

    if reexecution_options is not None and op_selection is not None:
        raise DagsterInvariantViolationError(
            "re-execution and op selection cannot be used together at this time."
        )

    if reexecution_options:
        if run_config is None:
            run = check.not_none(instance.get_run_by_id(reexecution_options.parent_run_id))
            run_config = run.run_config
        return _reexecute_job(
            job_arg=job_def,
            run_config=run_config,
            reexecution_options=reexecution_options,
            tags=tags,
            instance=instance,
            raise_on_error=raise_on_error,
        )
    else:
        return _logged_execute_job(
            job_arg=job_def,
            instance=instance,
            run_config=run_config,
            tags=tags,
            op_selection=op_selection,
            raise_on_error=raise_on_error,
            asset_selection=asset_selection,
        )


@telemetry_wrapper
def _logged_execute_job(
    job_arg: Union[IJob, JobDefinition],
    instance: DagsterInstance,
    run_config: Optional[Mapping[str, object]] = None,
    tags: Optional[Mapping[str, str]] = None,
    op_selection: Optional[Sequence[str]] = None,
    raise_on_error: bool = True,
    asset_selection: Optional[Sequence[AssetKey]] = None,
) -> JobExecutionResult:
    check.inst_param(instance, "instance", DagsterInstance)

    job_arg, repository_load_data = _job_with_repository_load_data(job_arg)

    (
        job_arg,
        run_config,
        tags,
        resolved_op_selection,
        op_selection,
    ) = _check_execute_job_args(
        job_arg=job_arg,
        run_config=run_config,
        tags=tags,
        op_selection=op_selection,
    )

    log_repo_stats(instance=instance, job=job_arg, source="execute_pipeline")

    dagster_run = instance.create_run_for_job(
        job_def=job_arg.get_definition(),
        run_config=run_config,
        op_selection=op_selection,
        resolved_op_selection=resolved_op_selection,
        tags=tags,
        job_code_origin=(
            job_arg.get_python_origin() if isinstance(job_arg, ReconstructableJob) else None
        ),
        repository_load_data=repository_load_data,
        asset_selection=frozenset(asset_selection) if asset_selection else None,
    )

    return execute_run(
        job_arg,
        dagster_run,
        instance,
        raise_on_error=raise_on_error,
    )


def _reexecute_job(
    job_arg: Union[IJob, JobDefinition],
    run_config: Optional[Mapping[str, object]],
    reexecution_options: ReexecutionOptions,
    tags: Optional[Mapping[str, str]],
    instance: DagsterInstance,
    raise_on_error: bool,
) -> JobExecutionResult:
    """Reexecute an existing job run."""
    with ephemeral_instance_if_missing(instance) as execute_instance:
        job_arg, repository_load_data = _job_with_repository_load_data(job_arg)

        (job_arg, run_config, tags, _, _) = _check_execute_job_args(
            job_arg=job_arg,
            run_config=run_config,
            tags=tags,
        )

        parent_dagster_run = execute_instance.get_run_by_id(reexecution_options.parent_run_id)
        if parent_dagster_run is None:
            check.failed(
                f"No parent run with id {reexecution_options.parent_run_id} found in instance.",
            )

        execution_plan: Optional[ExecutionPlan] = None
        # resolve step selection DSL queries using parent execution information
        if isinstance(reexecution_options, ReexecuteFromFailureOption):
            step_keys, known_state = KnownExecutionState.build_resume_retry_reexecution(
                instance=instance,
                parent_run=parent_dagster_run,
            )
            execution_plan = create_execution_plan(
                job_arg,
                run_config,
                step_keys_to_execute=step_keys,
                known_state=known_state,
                tags=parent_dagster_run.tags,
            )
        elif reexecution_options.step_selection:
            execution_plan = _resolve_reexecute_step_selection(
                execute_instance,
                job_arg,
                run_config,
                cast(DagsterRun, parent_dagster_run),
                reexecution_options.step_selection,
            )
        # else all steps will be executed and parent state is not needed

        if parent_dagster_run.asset_selection:
            job_arg = job_arg.get_subset(
                op_selection=None, asset_selection=parent_dagster_run.asset_selection
            )

        dagster_run = execute_instance.create_run_for_job(
            job_def=job_arg.get_definition(),
            execution_plan=execution_plan,
            run_config=run_config,
            tags=tags,
            op_selection=parent_dagster_run.op_selection,
            asset_selection=parent_dagster_run.asset_selection,
            resolved_op_selection=parent_dagster_run.resolved_op_selection,
            root_run_id=parent_dagster_run.root_run_id or parent_dagster_run.run_id,
            parent_run_id=parent_dagster_run.run_id,
            job_code_origin=(
                job_arg.get_python_origin() if isinstance(job_arg, ReconstructableJob) else None
            ),
            repository_load_data=repository_load_data,
        )

        return execute_run(
            job_arg,
            dagster_run,
            execute_instance,
            raise_on_error=raise_on_error,
        )


def execute_plan_iterator(
    execution_plan: ExecutionPlan,
    job: IJob,
    dagster_run: DagsterRun,
    instance: DagsterInstance,
    retry_mode: Optional[RetryMode] = None,
    run_config: Optional[Mapping[str, object]] = None,
) -> Iterator[DagsterEvent]:
    check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
    check.inst_param(job, "job", IJob)
    check.inst_param(dagster_run, "dagster_run", DagsterRun)
    check.inst_param(instance, "instance", DagsterInstance)
    retry_mode = check.opt_inst_param(retry_mode, "retry_mode", RetryMode, RetryMode.DISABLED)
    run_config = check.opt_mapping_param(run_config, "run_config")

    if isinstance(job, ReconstructableJob):
        job = job.with_repository_load_data(execution_plan.repository_load_data)

    return iter(
        ExecuteRunWithPlanIterable(
            execution_plan=execution_plan,
            iterator=inner_plan_execution_iterator,
            execution_context_manager=PlanExecutionContextManager(
                job=job,
                retry_mode=retry_mode,
                execution_plan=execution_plan,
                run_config=run_config,
                dagster_run=dagster_run,
                instance=instance,
            ),
        )
    )


def execute_plan(
    execution_plan: ExecutionPlan,
    job: IJob,
    instance: DagsterInstance,
    dagster_run: DagsterRun,
    run_config: Optional[Mapping[str, object]] = None,
    retry_mode: Optional[RetryMode] = None,
) -> Sequence[DagsterEvent]:
    """This is the entry point of dagster-graphql executions. For the dagster CLI entry point, see
    execute_job() above.
    """
    check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
    check.inst_param(job, "job", IJob)
    check.inst_param(instance, "instance", DagsterInstance)
    check.inst_param(dagster_run, "dagster_run", DagsterRun)
    run_config = check.opt_mapping_param(run_config, "run_config")
    check.opt_inst_param(retry_mode, "retry_mode", RetryMode)

    return list(
        execute_plan_iterator(
            execution_plan=execution_plan,
            job=job,
            run_config=run_config,
            dagster_run=dagster_run,
            instance=instance,
            retry_mode=retry_mode,
        )
    )


def _get_execution_plan_from_run(
    job: IJob,
    dagster_run: DagsterRun,
    instance: DagsterInstance,
) -> ExecutionPlan:
    execution_plan_snapshot = (
        instance.get_execution_plan_snapshot(dagster_run.execution_plan_snapshot_id)
        if dagster_run.execution_plan_snapshot_id
        else None
    )

    return create_execution_plan(
        job,
        run_config=dagster_run.run_config,
        step_keys_to_execute=dagster_run.step_keys_to_execute,
        instance_ref=instance.get_ref() if instance.is_persistent else None,
        tags=dagster_run.tags,
        repository_load_data=(
            execution_plan_snapshot.repository_load_data if execution_plan_snapshot else None
        ),
        known_state=(
            execution_plan_snapshot.initial_known_state if execution_plan_snapshot else None
        ),
    )


def create_execution_plan(
    job: Union[IJob, JobDefinition],
    run_config: Optional[Mapping[str, object]] = None,
    step_keys_to_execute: Optional[Sequence[str]] = None,
    known_state: Optional[KnownExecutionState] = None,
    instance_ref: Optional[InstanceRef] = None,
    tags: Optional[Mapping[str, str]] = None,
    repository_load_data: Optional[RepositoryLoadData] = None,
) -> ExecutionPlan:
    if isinstance(job, IJob):
        # If you have repository_load_data, make sure to use it when building plan
        if isinstance(job, ReconstructableJob) and repository_load_data is not None:
            job = job.with_repository_load_data(repository_load_data)
        job_def = job.get_definition()
    else:
        job_def = job

    run_config = check.opt_mapping_param(run_config, "run_config", key_type=str)
    check.opt_nullable_sequence_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
    check.opt_inst_param(instance_ref, "instance_ref", InstanceRef)
    tags = check.opt_mapping_param(tags, "tags", key_type=str, value_type=str)
    known_state = check.opt_inst_param(
        known_state,
        "known_state",
        KnownExecutionState,
        default=KnownExecutionState(),
    )
    repository_load_data = check.opt_inst_param(
        repository_load_data, "repository_load_data", RepositoryLoadData
    )

    resolved_run_config = ResolvedRunConfig.build(job_def, run_config)

    return ExecutionPlan.build(
        job_def,
        resolved_run_config,
        step_keys_to_execute=step_keys_to_execute,
        known_state=known_state,
        instance_ref=instance_ref,
        tags=tags,
        repository_load_data=repository_load_data,
    )


def job_execution_iterator(
    job_context: PlanOrchestrationContext, execution_plan: ExecutionPlan
) -> Iterator[DagsterEvent]:
    """A complete execution of a pipeline. Yields pipeline start, success,
    and failure events.

    Args:
        pipeline_context (PlanOrchestrationContext):
        execution_plan (ExecutionPlan):
    """
    # TODO: restart event?
    if not job_context.resume_from_failure:
        yield DagsterEvent.job_start(job_context)

    job_exception_info = None
    job_canceled_info = None
    failed_steps: list[
        DagsterEvent
    ] = []  # A list of failed steps, with the earliest failure event at the front
    generator_closed = False
    try:
        for event in job_context.executor.execute(job_context, execution_plan):
            if event.is_step_failure:
                failed_steps.append(event)
            elif event.is_resource_init_failure and event.step_key:
                failed_steps.append(event)

            # Telemetry
            log_dagster_event(event, job_context)

            yield event
    except GeneratorExit:
        # Shouldn't happen, but avoid runtime-exception in case this generator gets GC-ed
        # (see https://amir.rachum.com/blog/2017/03/03/generator-cleanup/).
        generator_closed = True
        job_exception_info = serializable_error_info_from_exc_info(sys.exc_info())
        if job_context.raise_on_error:
            raise
    except (KeyboardInterrupt, DagsterExecutionInterruptedError):
        job_canceled_info = serializable_error_info_from_exc_info(sys.exc_info())
        if job_context.raise_on_error:
            raise
    except BaseException:
        job_exception_info = serializable_error_info_from_exc_info(sys.exc_info())
        if job_context.raise_on_error:
            raise  # finally block will run before this is re-raised
    finally:
        if job_canceled_info:
            reloaded_run = job_context.instance.get_run_by_id(job_context.run_id)
            if reloaded_run and reloaded_run.status == DagsterRunStatus.CANCELING:
                event = DagsterEvent.job_canceled(job_context, job_canceled_info)
            elif reloaded_run and reloaded_run.status == DagsterRunStatus.CANCELED:
                # This happens if the run was force-terminated but was still able to send
                # a cancellation request
                event = DagsterEvent.engine_event(
                    job_context,
                    "Computational resources were cleaned up after the run was forcibly marked"
                    " as canceled.",
                    EngineEventData(),
                )
            elif job_context.instance.run_will_resume(job_context.run_id):
                event = DagsterEvent.engine_event(
                    job_context,
                    "Execution was interrupted unexpectedly. No user initiated termination"
                    " request was found, not treating as failure because run will be resumed.",
                    EngineEventData(),
                )
            elif reloaded_run and reloaded_run.status == DagsterRunStatus.FAILURE:
                event = DagsterEvent.engine_event(
                    job_context,
                    "Execution was interrupted for a run that was already in a failure state.",
                    EngineEventData(),
                )
            else:
                event = DagsterEvent.job_failure(
                    job_context,
                    "Execution was interrupted unexpectedly. "
                    "No user initiated termination request was found, treating as failure.",
                    failure_reason=RunFailureReason.UNEXPECTED_TERMINATION,
                    error_info=job_canceled_info,
                )
        elif job_exception_info:
            reloaded_run = job_context.instance.get_run_by_id(job_context.run_id)
            if reloaded_run and reloaded_run.status == DagsterRunStatus.CANCELING:
                event = DagsterEvent.job_canceled(
                    job_context,
                    error_info=job_exception_info,
                    message="Run failed after it was requested to be terminated.",
                )
            else:
                event = DagsterEvent.job_failure(
                    job_context,
                    "An exception was thrown during execution.",
                    failure_reason=RunFailureReason.RUN_EXCEPTION,
                    error_info=job_exception_info,
                )
        elif failed_steps:
            reloaded_run = job_context.instance.get_run_by_id(job_context.run_id)
            failed_step_keys = [event.step_key for event in failed_steps]

            if reloaded_run and reloaded_run.status == DagsterRunStatus.CANCELING:
                event = DagsterEvent.job_canceled(
                    job_context,
                    error_info=None,
                    message=f"Run was canceled. Failed steps: {failed_step_keys}.",
                )
            else:
                event = DagsterEvent.job_failure(
                    job_context,
                    f"Steps failed: {failed_step_keys}.",
                    failure_reason=RunFailureReason.STEP_FAILURE,
                    first_step_failure_event=failed_steps[0],
                )
        else:
            event = DagsterEvent.job_success(job_context)
        if not generator_closed:
            yield event


class ExecuteRunWithPlanIterable:
    """Utility class to consolidate execution logic.

    This is a class and not a function because, e.g., in constructing a `scoped_pipeline_context`
    for `JobExecutionResult`, we need to pull out the `pipeline_context` after we're done
    yielding events. This broadly follows a pattern we make use of in other places,
    cf. `dagster._utils.EventGenerationManager`.
    """

    def __init__(
        self,
        execution_plan: ExecutionPlan,
        iterator: Callable[..., Iterator[DagsterEvent]],
        execution_context_manager: ExecutionContextManager[Any],
    ):
        self.execution_plan = check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
        self.iterator = check.callable_param(iterator, "iterator")
        self.execution_context_manager = check.inst_param(
            execution_context_manager, "execution_context_manager", ExecutionContextManager
        )

        self.job_context = None

    def __iter__(self) -> Iterator[DagsterEvent]:
        # Since interrupts can't be raised at arbitrary points safely, delay them until designated
        # checkpoints during the execution.
        # To be maximally certain that interrupts are always caught during an execution process,
        # you can safely add an additional `with capture_interrupts()` at the very beginning of the
        # process that performs the execution.
        with capture_interrupts():
            yield from self.execution_context_manager.prepare_context()
            self.job_context = self.execution_context_manager.get_context()
            generator_closed = False
            try:
                if self.job_context:  # False if we had a pipeline init failure
                    yield from self.iterator(
                        execution_plan=self.execution_plan,
                        job_context=self.job_context,
                    )
            except GeneratorExit:
                # Shouldn't happen, but avoid runtime-exception in case this generator gets GC-ed
                # (see https://amir.rachum.com/blog/2017/03/03/generator-cleanup/).
                generator_closed = True
                raise
            finally:
                for event in self.execution_context_manager.shutdown_context():
                    if not generator_closed:
                        yield event


def _check_execute_job_args(
    job_arg: Union[JobDefinition, IJob],
    run_config: Optional[Mapping[str, object]],
    tags: Optional[Mapping[str, str]],
    op_selection: Optional[Sequence[str]] = None,
) -> tuple[
    IJob,
    Optional[Mapping],
    Mapping[str, str],
    Optional[AbstractSet[str]],
    Optional[Sequence[str]],
]:
    ijob = InMemoryJob(job_arg) if isinstance(job_arg, JobDefinition) else job_arg
    job_def = job_arg if isinstance(job_arg, JobDefinition) else job_arg.get_definition()

    run_config = check.opt_mapping_param(run_config, "run_config")

    tags = check.opt_mapping_param(tags, "tags", key_type=str)
    check.opt_sequence_param(op_selection, "op_selection", of_type=str)

    tags = merge_dicts(job_def.run_tags, tags)

    # generate job subset from the given op_selection
    if op_selection:
        ijob = ijob.get_subset(op_selection=op_selection)

    return (
        ijob,
        run_config,
        tags,
        ijob.resolved_op_selection,
        op_selection,
    )


def _resolve_reexecute_step_selection(
    instance: DagsterInstance,
    job: IJob,
    run_config: Optional[Mapping],
    parent_dagster_run: DagsterRun,
    step_selection: Sequence[str],
) -> ExecutionPlan:
    if parent_dagster_run.op_selection:
        job = job.get_subset(op_selection=parent_dagster_run.op_selection)

    state = KnownExecutionState.build_for_reexecution(instance, parent_dagster_run)

    parent_plan = create_execution_plan(
        job,
        parent_dagster_run.run_config,
        known_state=state,
    )
    step_keys_to_execute = parse_step_selection(parent_plan.get_all_step_deps(), step_selection)
    return create_execution_plan(
        job,
        run_config,
        step_keys_to_execute=list(step_keys_to_execute),
        known_state=state.update_for_step_selection(step_keys_to_execute),
        tags=parent_dagster_run.tags,
    )


def _job_with_repository_load_data(
    job_arg: Union[JobDefinition, IJob],
) -> tuple[Union[JobDefinition, IJob], Optional[RepositoryLoadData]]:
    """For ReconstructableJob, generate and return any required RepositoryLoadData, alongside
    a ReconstructableJob with this repository load data baked in.
    """
    if isinstance(job_arg, ReconstructableJob):
        # Unless this ReconstructableJob alread has repository_load_data attached, this will
        # force the repository_load_data to be computed from scratch.
        repository_load_data = job_arg.repository.get_definition().repository_load_data
        return job_arg.with_repository_load_data(repository_load_data), repository_load_data
    return job_arg, None



from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from enum import Enum
from typing import Any, Optional, cast

import dagster._check as check
from dagster._core.definitions import ExpectationResult
from dagster._core.events import (
    MARKER_EVENTS,
    PIPELINE_EVENTS,
    DagsterEventType,
    StepExpectationResultData,
)
from dagster._core.events.log import EventLogEntry
from dagster._core.storage.dagster_run import DagsterRunStatsSnapshot
from dagster._record import IHaveNew, record, record_custom
from dagster._serdes import whitelist_for_serdes

RUN_STATS_EVENT_TYPES = {
    *PIPELINE_EVENTS,
    DagsterEventType.STEP_FAILURE,
    DagsterEventType.STEP_SUCCESS,
    DagsterEventType.ASSET_MATERIALIZATION,
    DagsterEventType.STEP_EXPECTATION_RESULT,
}

STEP_STATS_EVENT_TYPES = {
    DagsterEventType.STEP_START,
    DagsterEventType.STEP_FAILURE,
    DagsterEventType.STEP_RESTARTED,
    DagsterEventType.STEP_SUCCESS,
    DagsterEventType.STEP_SKIPPED,
    DagsterEventType.ASSET_MATERIALIZATION,
    DagsterEventType.STEP_EXPECTATION_RESULT,
    DagsterEventType.STEP_UP_FOR_RETRY,
    DagsterEventType.STEP_RESTARTED,
    *MARKER_EVENTS,
}


def build_run_stats_from_events(
    run_id: str,
    entries: Iterable[EventLogEntry],
    previous_stats: Optional[DagsterRunStatsSnapshot] = None,
) -> DagsterRunStatsSnapshot:
    try:
        iter(entries)
    except TypeError as exc:
        raise check.ParameterCheckError(
            "Invariant violation for parameter 'records'. Description: Expected iterable."
        ) from exc
    for i, entry in enumerate(entries):
        check.inst_param(entry, f"entries[{i}]", EventLogEntry)

    if previous_stats:
        steps_succeeded = previous_stats.steps_succeeded
        steps_failed = previous_stats.steps_failed
        materializations = previous_stats.materializations
        expectations = previous_stats.expectations
        enqueued_time = previous_stats.enqueued_time
        launch_time = previous_stats.launch_time
        start_time = previous_stats.start_time
        end_time = previous_stats.end_time
    else:
        steps_succeeded = 0
        steps_failed = 0
        materializations = 0
        expectations = 0
        enqueued_time = None
        launch_time = None
        start_time = None
        end_time = None

    for event in entries:
        if not event.is_dagster_event:
            continue
        dagster_event = event.get_dagster_event()

        if dagster_event.event_type == DagsterEventType.PIPELINE_START:
            start_time = event.timestamp
        if dagster_event.event_type == DagsterEventType.PIPELINE_STARTING:
            launch_time = event.timestamp
        if dagster_event.event_type == DagsterEventType.PIPELINE_ENQUEUED:
            enqueued_time = event.timestamp
        if dagster_event.event_type == DagsterEventType.STEP_FAILURE:
            steps_failed += 1
        if dagster_event.event_type == DagsterEventType.STEP_SUCCESS:
            steps_succeeded += 1
        if dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION:
            materializations += 1
        if dagster_event.event_type == DagsterEventType.STEP_EXPECTATION_RESULT:
            expectations += 1
        if (
            dagster_event.event_type == DagsterEventType.PIPELINE_SUCCESS
            or dagster_event.event_type == DagsterEventType.PIPELINE_FAILURE
            or dagster_event.event_type == DagsterEventType.PIPELINE_CANCELED
        ):
            end_time = event.timestamp

    return DagsterRunStatsSnapshot(
        run_id,
        steps_succeeded,
        steps_failed,
        materializations,
        expectations,
        enqueued_time,
        launch_time,
        start_time,
        end_time,
    )


@whitelist_for_serdes
class StepEventStatus(Enum):
    SKIPPED = "SKIPPED"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    IN_PROGRESS = "IN_PROGRESS"


@whitelist_for_serdes
@record_custom
class RunStepMarker(IHaveNew):
    start_time: Optional[float]
    end_time: Optional[float]
    key: Optional[str]

    def __new__(
        cls,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        key: Optional[str] = None,
    ):
        return super().__new__(
            cls,
            start_time=check.opt_float_param(start_time, "start_time"),
            end_time=check.opt_float_param(end_time, "end_time"),
            key=check.opt_str_param(key, "key"),
        )


@whitelist_for_serdes
@record_custom
class RunStepKeyStatsSnapshot(IHaveNew):
    run_id: str
    step_key: str
    status: Optional[StepEventStatus]
    start_time: Optional[float]
    end_time: Optional[float]
    materialization_events: Sequence[EventLogEntry]
    expectation_results: Sequence[ExpectationResult]
    attempts: Optional[int]
    attempts_list: Sequence[RunStepMarker]
    markers: Sequence[RunStepMarker]
    partial_attempt_start: Optional[float]

    def __new__(
        cls,
        run_id: str,
        step_key: str,
        status: Optional[StepEventStatus] = None,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        materialization_events: Optional[Sequence[EventLogEntry]] = None,
        expectation_results: Optional[Sequence[ExpectationResult]] = None,
        attempts: Optional[int] = None,
        attempts_list: Optional[Sequence[RunStepMarker]] = None,
        markers: Optional[Sequence[RunStepMarker]] = None,
        partial_attempt_start: Optional[float] = None,
    ):
        return super().__new__(
            cls,
            run_id=check.str_param(run_id, "run_id"),
            step_key=check.str_param(step_key, "step_key"),
            status=check.opt_inst_param(status, "status", StepEventStatus),
            start_time=check.opt_float_param(start_time, "start_time"),
            end_time=check.opt_float_param(end_time, "end_time"),
            materialization_events=check.opt_sequence_param(
                materialization_events,
                "materialization_events",
                EventLogEntry,
            ),
            expectation_results=check.opt_sequence_param(
                expectation_results, "expectation_results", ExpectationResult
            ),
            attempts=check.opt_int_param(attempts, "attempts"),
            attempts_list=check.opt_sequence_param(attempts_list, "attempts_list", RunStepMarker),
            markers=check.opt_sequence_param(markers, "markers", RunStepMarker),
            # used to calculate incremental step stats using batches of event logs
            partial_attempt_start=check.opt_float_param(
                partial_attempt_start, "partial_attempt_start"
            ),
        )


@whitelist_for_serdes
@record
class RunStepStatsSnapshot:
    run_id: str
    step_key_stats: Sequence[RunStepKeyStatsSnapshot]
    partial_markers: Optional[Mapping[str, Sequence[RunStepMarker]]]


def build_run_step_stats_from_events(
    run_id: str,
    entries: Iterable[EventLogEntry],
) -> Sequence[RunStepKeyStatsSnapshot]:
    snapshot = build_run_step_stats_snapshot_from_events(run_id, entries)
    return snapshot.step_key_stats


def build_run_step_stats_snapshot_from_events(
    run_id: str,
    entries: Iterable[EventLogEntry],
    previous_snapshot: Optional["RunStepStatsSnapshot"] = None,
) -> "RunStepStatsSnapshot":
    by_step_key: dict[str, dict[str, Any]] = defaultdict(dict)
    attempts = defaultdict(list)
    markers: dict[str, dict[str, Any]] = defaultdict(dict)

    if previous_snapshot:
        for step_stats in previous_snapshot.step_key_stats:
            check.invariant(step_stats.run_id == run_id)
            by_step_key[step_stats.step_key] = {
                "start_time": step_stats.start_time,
                "end_time": step_stats.end_time,
                "status": step_stats.status,
                "materialization_events": step_stats.materialization_events,
                "expectation_results": step_stats.expectation_results,
                "attempts": step_stats.attempts,
                "partial_attempt_start": step_stats.partial_attempt_start,
            }
            for attempt in step_stats.attempts_list:
                attempts[step_stats.step_key].append(attempt)

            for marker in step_stats.markers:
                assert marker.key
                markers[step_stats.step_key][marker.key] = {
                    "key": marker.key,
                    "start": marker.start_time,
                    "end": marker.end_time,
                }

        # handle the partial markers
        if previous_snapshot.partial_markers:
            for step_key, partial_markers in previous_snapshot.partial_markers.items():
                for marker in partial_markers:
                    assert marker.key
                    markers[step_key][marker.key] = {
                        "key": marker.key,
                        "start": marker.start_time,
                        "end": marker.end_time,
                    }

    def _open_attempt(step_key: str, event: EventLogEntry) -> None:
        by_step_key[step_key]["attempts"] = int(by_step_key[step_key].get("attempts") or 0) + 1
        by_step_key[step_key]["partial_attempt_start"] = event.timestamp

    def _close_attempt(step_key: str, event: EventLogEntry) -> None:
        start_time = by_step_key[step_key].get("partial_attempt_start")

        if start_time is None:
            # this should only happen if the step was retried before starting (weird)
            by_step_key[step_key]["attempts"] = int(by_step_key[step_key].get("attempts") or 0) + 1
            start_time = event.timestamp

        attempts[step_key].append(
            RunStepMarker(
                start_time=start_time,
                end_time=event.timestamp,
            )
        )
        by_step_key[step_key]["partial_attempt_start"] = None

    for event in entries:
        if not event.is_dagster_event:
            continue
        dagster_event = event.get_dagster_event()

        step_key = dagster_event.step_key
        if not step_key:
            continue

        if dagster_event.event_type not in STEP_STATS_EVENT_TYPES:
            continue

        if dagster_event.event_type == DagsterEventType.STEP_START:
            by_step_key[step_key]["status"] = StepEventStatus.IN_PROGRESS
            by_step_key[step_key]["start_time"] = event.timestamp
            _open_attempt(step_key, event)
        if dagster_event.event_type == DagsterEventType.STEP_RESTARTED:
            _open_attempt(step_key, event)
        if dagster_event.event_type == DagsterEventType.STEP_UP_FOR_RETRY:
            _close_attempt(step_key, event)
        if dagster_event.event_type == DagsterEventType.STEP_FAILURE:
            by_step_key[step_key]["end_time"] = event.timestamp
            by_step_key[step_key]["status"] = StepEventStatus.FAILURE
            _close_attempt(step_key, event)
        if dagster_event.event_type == DagsterEventType.STEP_SUCCESS:
            by_step_key[step_key]["end_time"] = event.timestamp
            by_step_key[step_key]["status"] = StepEventStatus.SUCCESS
            _close_attempt(step_key, event)
        if dagster_event.event_type == DagsterEventType.STEP_SKIPPED:
            by_step_key[step_key]["end_time"] = event.timestamp
            by_step_key[step_key]["status"] = StepEventStatus.SKIPPED
            _close_attempt(step_key, event)
        if dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION:
            materialization_events = by_step_key[step_key].get("materialization_events", [])
            materialization_events.append(event)
            by_step_key[step_key]["materialization_events"] = materialization_events
        if dagster_event.event_type == DagsterEventType.STEP_EXPECTATION_RESULT:
            expectation_data = cast(StepExpectationResultData, dagster_event.event_specific_data)
            expectation_result = expectation_data.expectation_result
            step_expectation_results = by_step_key[step_key].get("expectation_results", [])
            step_expectation_results.append(expectation_result)
            by_step_key[step_key]["expectation_results"] = step_expectation_results

        if dagster_event.event_type in MARKER_EVENTS:
            if dagster_event.engine_event_data.marker_start:
                marker_key = dagster_event.engine_event_data.marker_start
                if marker_key not in markers[step_key]:
                    markers[step_key][marker_key] = {"key": marker_key, "start": event.timestamp}
                else:
                    markers[step_key][marker_key]["start"] = event.timestamp

            if dagster_event.engine_event_data.marker_end:
                marker_key = dagster_event.engine_event_data.marker_end
                if marker_key not in markers[step_key]:
                    markers[step_key][marker_key] = {"key": marker_key, "end": event.timestamp}
                else:
                    markers[step_key][marker_key]["end"] = event.timestamp

    snapshots = []
    for step_key, step_stats in by_step_key.items():
        snapshots.append(
            RunStepKeyStatsSnapshot(
                run_id=run_id,
                step_key=step_key,
                **step_stats,
                markers=[
                    RunStepMarker(
                        start_time=marker.get("start"),
                        end_time=marker.get("end"),
                        key=marker.get("key"),
                    )
                    for marker in markers[step_key].values()
                ],
                attempts_list=attempts[step_key],
            )
        )

    return RunStepStatsSnapshot(
        run_id=run_id,
        step_key_stats=snapshots,
        partial_markers={
            step_key: [
                RunStepMarker(start_time=marker.get("start"), end_time=marker.get("end"), key=key)
                for key, marker in markers.items()
            ]
            for step_key, markers in markers.items()
            if step_key not in by_step_key
        },
    )



import base64
import dataclasses
import datetime
import logging
import sys
import threading
import zlib
from collections import defaultdict
from collections.abc import Mapping, Sequence
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import ExitStack
from types import TracebackType
from typing import AbstractSet, Any, Optional, cast  # noqa: UP035

import dagster._check as check
from dagster._core.definitions.asset_daemon_cursor import (
    AssetDaemonCursor,
    LegacyAssetDaemonCursorWrapper,
    backcompat_deserialize_asset_daemon_cursor_str,
)
from dagster._core.definitions.asset_key import AssetCheckKey, EntityKey
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.automation_condition_sensor_definition import (
    EMIT_BACKFILLS_METADATA_KEY,
)
from dagster._core.definitions.automation_tick_evaluation_context import (
    AutomationTickEvaluationContext,
)
from dagster._core.definitions.base_asset_graph import BaseAssetGraph
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionEvaluationWithRunIds,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.remote_asset_graph import RemoteAssetGraph
from dagster._core.definitions.repository_definition.valid_definitions import (
    SINGLETON_REPOSITORY_NAME,
)
from dagster._core.definitions.run_request import InstigatorType, RunRequest
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.definitions.sensor_definition import DefaultSensorStatus
from dagster._core.errors import DagsterCodeLocationLoadError, DagsterUserCodeUnreachableError
from dagster._core.execution.backfill import PartitionBackfill
from dagster._core.execution.submit_asset_runs import RunRequestExecutionData, submit_asset_run
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation import RemoteSensor
from dagster._core.remote_representation.external import RemoteRepository
from dagster._core.remote_representation.origin import RemoteInstigatorOrigin
from dagster._core.scheduler.instigation import (
    InstigatorState,
    InstigatorStatus,
    InstigatorTick,
    SensorInstigatorData,
    TickData,
    TickStatus,
)
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.storage.tags import (
    ASSET_EVALUATION_ID_TAG,
    AUTO_MATERIALIZE_TAG,
    AUTO_OBSERVE_TAG,
    AUTOMATION_CONDITION_TAG,
    SENSOR_NAME_TAG,
)
from dagster._core.utils import (
    InheritContextThreadPoolExecutor,
    make_new_backfill_id,
    make_new_run_id,
)
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._daemon.daemon import DaemonIterator, DagsterDaemon, SpanMarker
from dagster._daemon.sensor import is_under_min_interval, mark_sensor_state_for_tick
from dagster._daemon.utils import DaemonErrorCapture
from dagster._serdes import serialize_value
from dagster._serdes.serdes import deserialize_value
from dagster._time import get_current_datetime, get_current_timestamp
from dagster._utils import SingleInstigatorDebugCrashFlags, check_for_debug_crash, return_as_list

_LEGACY_PRE_SENSOR_AUTO_MATERIALIZE_CURSOR_KEY = "ASSET_DAEMON_CURSOR"
_PRE_SENSOR_AUTO_MATERIALIZE_CURSOR_KEY = "ASSET_DAEMON_CURSOR_NEW"
_PRE_SENSOR_ASSET_DAEMON_PAUSED_KEY = "ASSET_DAEMON_PAUSED"
_MIGRATED_CURSOR_TO_SENSORS_KEY = "MIGRATED_CURSOR_TO_SENSORS"
_MIGRATED_SENSOR_NAMES_KEY = "MIGRATED_SENSOR_NAMES_KEY"


EVALUATIONS_TTL_DAYS = 30

# When retrying a tick, how long to wait before ignoring it and moving on to the next one
# (To account for the rare case where the daemon is down for a long time, starts back up, and
# there's an old in-progress tick left to finish that may no longer be correct to finish)
MAX_TIME_TO_RESUME_TICK_SECONDS = 60 * 60 * 24

_PRE_SENSOR_AUTO_MATERIALIZE_ORIGIN_ID = "asset_daemon_origin"
_PRE_SENSOR_AUTO_MATERIALIZE_SELECTOR_ID = "asset_daemon_selector"
_PRE_SENSOR_AUTO_MATERIALIZE_INSTIGATOR_NAME = "asset_daemon"

MIN_INTERVAL_LOOP_SECONDS = 5


def _get_has_migrated(instance: DagsterInstance, migration_key: str) -> bool:
    return bool(
        instance.daemon_cursor_storage.get_cursor_values({migration_key}).get(migration_key)
    )


def _set_has_migrated(instance: DagsterInstance, migration_key: str) -> None:
    instance.daemon_cursor_storage.set_cursor_values({migration_key: "1"})


def get_has_migrated_to_sensors(instance: DagsterInstance) -> bool:
    return _get_has_migrated(instance, _MIGRATED_CURSOR_TO_SENSORS_KEY)


def set_has_migrated_to_sensors(instance: DagsterInstance) -> None:
    _set_has_migrated(instance, _MIGRATED_CURSOR_TO_SENSORS_KEY)


def get_has_migrated_sensor_names(instance: DagsterInstance) -> bool:
    return _get_has_migrated(instance, _MIGRATED_SENSOR_NAMES_KEY)


def set_has_migrated_sensor_names(instance: DagsterInstance) -> None:
    _set_has_migrated(instance, _MIGRATED_SENSOR_NAMES_KEY)


def get_auto_materialize_paused(instance: DagsterInstance) -> bool:
    return (
        instance.daemon_cursor_storage.get_cursor_values({_PRE_SENSOR_ASSET_DAEMON_PAUSED_KEY}).get(
            _PRE_SENSOR_ASSET_DAEMON_PAUSED_KEY
        )
        != "false"
    )


def set_auto_materialize_paused(instance: DagsterInstance, paused: bool):
    instance.daemon_cursor_storage.set_cursor_values(
        {_PRE_SENSOR_ASSET_DAEMON_PAUSED_KEY: "true" if paused else "false"}
    )


def _get_pre_sensor_auto_materialize_cursor(
    instance: DagsterInstance, full_asset_graph: Optional[BaseAssetGraph]
) -> AssetDaemonCursor:
    """Gets a deserialized cursor by either reading from the new cursor key and simply deserializing
    the value, or by reading from the old cursor key and converting the legacy cursor into the
    updated format.
    """
    serialized_cursor = instance.daemon_cursor_storage.get_cursor_values(
        {_PRE_SENSOR_AUTO_MATERIALIZE_CURSOR_KEY}
    ).get(_PRE_SENSOR_AUTO_MATERIALIZE_CURSOR_KEY)

    if not serialized_cursor:
        # check for legacy cursor
        legacy_serialized_cursor = instance.daemon_cursor_storage.get_cursor_values(
            {_LEGACY_PRE_SENSOR_AUTO_MATERIALIZE_CURSOR_KEY}
        ).get(_LEGACY_PRE_SENSOR_AUTO_MATERIALIZE_CURSOR_KEY)
        return (
            backcompat_deserialize_asset_daemon_cursor_str(
                legacy_serialized_cursor, full_asset_graph, 0
            )
            if legacy_serialized_cursor
            else AssetDaemonCursor.empty()
        )
    else:
        return deserialize_value(serialized_cursor, AssetDaemonCursor)


def get_current_evaluation_id(
    instance: DagsterInstance, sensor_origin: Optional[RemoteInstigatorOrigin]
) -> Optional[int]:
    if not sensor_origin:
        cursor = _get_pre_sensor_auto_materialize_cursor(instance, None)
    else:
        instigator_state = check.not_none(instance.schedule_storage).get_instigator_state(
            sensor_origin.get_id(), sensor_origin.get_selector().get_id()
        )
        serialized_cursor = (
            cast(SensorInstigatorData, instigator_state.instigator_data).cursor
            if instigator_state
            else None
        )
        if not serialized_cursor:
            return None
        cursor = asset_daemon_cursor_from_instigator_serialized_cursor(serialized_cursor, None)

    return cursor.evaluation_id


def asset_daemon_cursor_to_instigator_serialized_cursor(cursor: AssetDaemonCursor) -> str:
    """This method compresses the serialized cursor and returns a b64 encoded string to be stored
    as a string value.
    """
    # increment the version if the cursor format changes
    VERSION = "0"

    serialized_bytes = serialize_value(cursor).encode("utf-8")
    compressed_bytes = zlib.compress(serialized_bytes)
    encoded_cursor = base64.b64encode(compressed_bytes).decode("utf-8")
    return VERSION + encoded_cursor


def asset_daemon_cursor_from_instigator_serialized_cursor(
    serialized_cursor: Optional[str], asset_graph: Optional[BaseAssetGraph]
) -> AssetDaemonCursor:
    """This method decompresses the serialized cursor and returns a deserialized cursor object,
    converting from the legacy cursor format if necessary.
    """
    if serialized_cursor is None:
        return AssetDaemonCursor.empty()

    version, encoded_bytes = serialized_cursor[0], serialized_cursor[1:]
    if version != "0":
        return AssetDaemonCursor.empty()

    decoded_bytes = base64.b64decode(encoded_bytes)
    decompressed_bytes = zlib.decompress(decoded_bytes)
    decompressed_str = decompressed_bytes.decode("utf-8")

    deserialized_cursor = deserialize_value(
        decompressed_str, (LegacyAssetDaemonCursorWrapper, AssetDaemonCursor)
    )
    if isinstance(deserialized_cursor, LegacyAssetDaemonCursorWrapper):
        return deserialized_cursor.get_asset_daemon_cursor(asset_graph)
    return deserialized_cursor


class AutoMaterializeLaunchContext:
    def __init__(
        self,
        tick: InstigatorTick,
        remote_sensor: Optional[RemoteSensor],
        instance: DagsterInstance,
        logger: logging.Logger,
        tick_retention_settings,
    ):
        self._tick = tick
        self._logger = logger
        self._instance = instance
        self._remote_sensor = remote_sensor

        self._purge_settings = defaultdict(set)
        for status, day_offset in tick_retention_settings.items():
            self._purge_settings[day_offset].add(status)

    @property
    def status(self) -> TickStatus:
        return self._tick.status

    @property
    def tick(self) -> InstigatorTick:
        return self._tick

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    def add_run_info(self, run_id=None):
        self._tick = self._tick.with_run_info(run_id)

    def set_run_requests(
        self,
        run_requests: Sequence[RunRequest],
        reserved_run_ids: Optional[Sequence[str]],
    ):
        self._tick = self._tick.with_run_requests(run_requests, reserved_run_ids=reserved_run_ids)
        return self._tick

    def update_state(self, status: TickStatus, **kwargs: object):
        if status in {TickStatus.SKIPPED, TickStatus.SUCCESS}:
            kwargs["failure_count"] = 0
            kwargs["consecutive_failure_count"] = 0

        self._tick = self._tick.with_status(status=status, **kwargs)

    def set_user_interrupted(self, user_interrupted: bool):
        self._tick = self._tick.with_user_interrupted(user_interrupted)

    def set_skip_reason(self, skip_reason: str):
        self._tick = self._tick.with_reason(skip_reason)

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type: type[BaseException],
        exception_value: Exception,
        traceback: TracebackType,
    ) -> None:
        if exception_value and isinstance(exception_value, KeyboardInterrupt):
            return

        # Log the error if the failure wasn't an interrupt or the daemon generator stopping
        if exception_value and not isinstance(exception_value, GeneratorExit):
            if isinstance(
                exception_value, (DagsterUserCodeUnreachableError, DagsterCodeLocationLoadError)
            ):
                try:
                    raise DagsterUserCodeUnreachableError(
                        "Unable to reach the code server. Automation condition evaluation will resume once the code server is available."
                    ) from exception_value
                except:
                    error_data = DaemonErrorCapture.process_exception(
                        sys.exc_info(),
                        logger=self.logger,
                        log_message="Asset daemon tick caught an error",
                    )
                    self.update_state(
                        TickStatus.FAILURE,
                        error=error_data,
                        # don't increment the failure count - retry until the server is available again
                        failure_count=self._tick.failure_count,
                        consecutive_failure_count=(
                            (self._tick.consecutive_failure_count or self._tick.failure_count) + 1
                        ),
                    )
            else:
                error_data = DaemonErrorCapture.process_exception(
                    sys.exc_info(),
                    logger=self.logger,
                    log_message="Asset daemon tick caught an error",
                )
                self.update_state(
                    TickStatus.FAILURE,
                    error=error_data,
                    failure_count=self._tick.failure_count + 1,
                    consecutive_failure_count=(
                        (self._tick.consecutive_failure_count or self._tick.failure_count) + 1
                    ),
                )

        check.invariant(
            self._tick.status != TickStatus.STARTED,
            "Tick must be in a terminal state when the AutoMaterializeLaunchContext is closed",
        )

        # write the new tick status to the database

        self.write()

        for day_offset, statuses in self._purge_settings.items():
            if day_offset <= 0:
                continue
            self._instance.purge_ticks(
                (
                    self._remote_sensor.get_remote_origin().get_id()
                    if self._remote_sensor
                    else _PRE_SENSOR_AUTO_MATERIALIZE_ORIGIN_ID
                ),
                (
                    self._remote_sensor.selector.get_id()
                    if self._remote_sensor
                    else _PRE_SENSOR_AUTO_MATERIALIZE_SELECTOR_ID
                ),
                before=(get_current_datetime() - datetime.timedelta(days=day_offset)).timestamp(),
                tick_statuses=list(statuses),
            )

    def write(self) -> None:
        self._instance.update_tick(self._tick)


class AssetDaemon(DagsterDaemon):
    def __init__(self, settings: Mapping[str, Any], pre_sensor_interval_seconds: int):
        self._pre_sensor_interval_seconds = pre_sensor_interval_seconds
        self._last_pre_sensor_submit_time = None

        self._checked_migrations = False

        self._settings = settings

        super().__init__()

    @classmethod
    def daemon_type(cls) -> str:
        return "ASSET"

    def _get_print_sensor_name(self, sensor: Optional[RemoteSensor]) -> str:
        if not sensor:
            return ""
        repo_origin = sensor.get_remote_origin().repository_origin
        repo_name = repo_origin.repository_name
        location_name = repo_origin.code_location_origin.location_name
        repo_name = (
            location_name
            if repo_name == SINGLETON_REPOSITORY_NAME
            else f"{repo_name}@{location_name}"
        )
        return f" for {sensor.name} in {repo_name}"

    def core_loop(
        self,
        workspace_process_context: IWorkspaceProcessContext,
        shutdown_event: threading.Event,
    ) -> DaemonIterator:
        instance: DagsterInstance = workspace_process_context.instance

        schedule_storage = check.not_none(
            instance.schedule_storage,
            "Auto materialization requires schedule storage to be configured",
        )

        if not schedule_storage.supports_auto_materialize_asset_evaluations:
            self._logger.warning(
                "Auto materialize evaluations are not getting logged. Run `dagster instance"
                " migrate` to enable."
            )

        amp_tick_futures: dict[Optional[str], Future] = {}
        threadpool_executor = None
        submit_threadpool_executor = None
        with ExitStack() as stack:
            if self._settings.get("use_threads"):
                threadpool_executor = stack.enter_context(
                    InheritContextThreadPoolExecutor(
                        max_workers=self._settings.get("num_workers"),
                        thread_name_prefix="asset_daemon_worker",
                    )
                )
            num_submit_workers = self._settings.get("num_submit_workers")
            if num_submit_workers:
                submit_threadpool_executor = stack.enter_context(
                    InheritContextThreadPoolExecutor(
                        max_workers=num_submit_workers,
                        thread_name_prefix="asset_daemon_submit_worker",
                    )
                )

            while True:
                start_time = get_current_timestamp()
                yield SpanMarker.START_SPAN
                try:
                    yield from self._run_iteration_impl(
                        workspace_process_context,
                        threadpool_executor=threadpool_executor,
                        submit_threadpool_executor=submit_threadpool_executor,
                        amp_tick_futures=amp_tick_futures,
                        debug_crash_flags={},
                    )
                except Exception:
                    error_info = DaemonErrorCapture.process_exception(
                        exc_info=sys.exc_info(),
                        logger=self._logger,
                        log_message="AssetDaemon caught an error",
                    )
                    yield error_info
                yield SpanMarker.END_SPAN
                end_time = get_current_timestamp()
                loop_duration = end_time - start_time
                sleep_time = max(0, MIN_INTERVAL_LOOP_SECONDS - loop_duration)
                shutdown_event.wait(sleep_time)
                yield None

    def _run_iteration_impl(
        self,
        workspace_process_context: IWorkspaceProcessContext,
        threadpool_executor: Optional[ThreadPoolExecutor],
        submit_threadpool_executor: Optional[ThreadPoolExecutor],
        amp_tick_futures: dict[Optional[str], Future],
        debug_crash_flags: SingleInstigatorDebugCrashFlags,
    ):
        instance: DagsterInstance = workspace_process_context.instance

        use_auto_materialize_sensors = instance.auto_materialize_use_sensors
        if get_auto_materialize_paused(instance) and not use_auto_materialize_sensors:
            yield
            return

        now = get_current_timestamp()

        workspace = workspace_process_context.create_request_context()

        sensors_and_repos: Sequence[tuple[Optional[RemoteSensor], Optional[RemoteRepository]]] = []

        if use_auto_materialize_sensors:
            current_workspace = {
                location_entry.origin.location_name: location_entry
                for location_entry in workspace.get_code_location_entries().values()
            }

            eligible_sensors_and_repos = []
            for location_entry in current_workspace.values():
                code_location = location_entry.code_location
                if code_location:
                    for repo in code_location.get_repositories().values():
                        for sensor in repo.get_sensors():
                            if sensor.sensor_type.is_handled_by_asset_daemon:
                                eligible_sensors_and_repos.append((sensor, repo))

            if not eligible_sensors_and_repos:
                return

            all_sensor_states = {
                sensor_state.selector_id: sensor_state
                for sensor_state in instance.all_instigator_state(
                    instigator_type=InstigatorType.SENSOR
                )
            }

            if not self._checked_migrations:
                if not get_has_migrated_to_sensors(instance):
                    # Do a one-time migration to create the cursors for each sensor, based on the
                    # existing cursor for the legacy AMP tick
                    asset_graph = workspace.asset_graph
                    pre_sensor_cursor = _get_pre_sensor_auto_materialize_cursor(
                        instance, asset_graph
                    )
                    if pre_sensor_cursor != AssetDaemonCursor.empty():
                        self._logger.info(
                            "Translating legacy cursor into a new cursor for each new automation policy sensor"
                        )
                        all_sensor_states = self._create_initial_sensor_cursors_from_raw_cursor(
                            instance,
                            eligible_sensors_and_repos,
                            all_sensor_states,
                            pre_sensor_cursor,
                        )

                    set_has_migrated_to_sensors(instance)
                if not get_has_migrated_sensor_names(instance):
                    # Do a one-time migration to copy state from sensors with the legacy default
                    # name to the new default name
                    self._logger.info(
                        "Renaming any states corresponding to the legacy default name"
                    )
                    all_sensor_states = self._copy_default_auto_materialize_sensor_states(
                        instance, all_sensor_states
                    )
                    set_has_migrated_sensor_names(instance)

                self._checked_migrations = True

            for sensor, repo in eligible_sensors_and_repos:
                selector_id = sensor.selector_id
                if sensor.get_current_instigator_state(
                    all_sensor_states.get(selector_id)
                ).is_running:
                    sensors_and_repos.append((sensor, repo))

        else:
            sensors_and_repos.append(
                (
                    None,
                    None,
                )  # Represents that there's a single set of ticks with no underlying sensor
            )
            all_sensor_states = {}

        for sensor, repo in sensors_and_repos:
            if sensor:
                selector_id = sensor.selector.get_id()
                auto_materialize_state = all_sensor_states.get(selector_id)
            else:
                selector_id = None
                auto_materialize_state = None

            if not sensor:
                # make sure we are only running every pre_sensor_interval_seconds
                if (
                    self._last_pre_sensor_submit_time
                    and now - self._last_pre_sensor_submit_time < self._pre_sensor_interval_seconds
                ):
                    continue

                self._last_pre_sensor_submit_time = now

            elif not auto_materialize_state:
                assert sensor.default_status == DefaultSensorStatus.RUNNING
                auto_materialize_state = InstigatorState(
                    sensor.get_remote_origin(),
                    InstigatorType.SENSOR,
                    InstigatorStatus.DECLARED_IN_CODE,
                    SensorInstigatorData(
                        min_interval=sensor.min_interval_seconds,
                        cursor=None,
                        last_sensor_start_timestamp=get_current_timestamp(),
                        sensor_type=sensor.sensor_type,
                    ),
                )
                instance.add_instigator_state(auto_materialize_state)
            elif is_under_min_interval(auto_materialize_state, sensor):
                continue

            if threadpool_executor:
                # only one tick per sensor can be in flight
                if selector_id in amp_tick_futures and not amp_tick_futures[selector_id].done():
                    continue

                future = threadpool_executor.submit(
                    self._process_auto_materialize_tick,
                    workspace_process_context,
                    repo,
                    sensor,
                    debug_crash_flags,
                    submit_threadpool_executor,
                )
                amp_tick_futures[selector_id] = future
                yield
            else:
                yield from self._process_auto_materialize_tick_generator(
                    workspace_process_context,
                    repo,
                    sensor,
                    debug_crash_flags,
                    submit_threadpool_executor,
                )

    def _create_initial_sensor_cursors_from_raw_cursor(
        self,
        instance: DagsterInstance,
        sensors_and_repos: Sequence[tuple[RemoteSensor, RemoteRepository]],
        all_sensor_states: Mapping[str, InstigatorState],
        pre_sensor_cursor: AssetDaemonCursor,
    ) -> Mapping[str, InstigatorState]:
        start_status = (
            InstigatorStatus.STOPPED
            if get_auto_materialize_paused(instance)
            else InstigatorStatus.RUNNING
        )

        result = {}

        for sensor, repo in sensors_and_repos:
            new_auto_materialize_state = InstigatorState(
                sensor.get_remote_origin(),
                InstigatorType.SENSOR,
                (
                    InstigatorStatus.DECLARED_IN_CODE
                    if sensor.default_status == DefaultSensorStatus.RUNNING
                    else start_status
                ),
                SensorInstigatorData(
                    min_interval=sensor.min_interval_seconds,
                    cursor=asset_daemon_cursor_to_instigator_serialized_cursor(pre_sensor_cursor),
                    last_sensor_start_timestamp=get_current_timestamp(),
                    sensor_type=sensor.sensor_type,
                ),
            )

            if all_sensor_states.get(sensor.selector_id):
                instance.update_instigator_state(new_auto_materialize_state)
            else:
                instance.add_instigator_state(new_auto_materialize_state)

            result[sensor.selector_id] = new_auto_materialize_state

        return result

    def _copy_default_auto_materialize_sensor_states(
        self,
        instance: DagsterInstance,
        all_sensor_states: Mapping[str, InstigatorState],
    ) -> Mapping[str, InstigatorState]:
        """Searches for sensors named `default_auto_materialize_sensor` and copies their state
        to a sensor in the same repository named `default_automation_condition_sensor`.
        """
        result = dict(all_sensor_states)

        for instigator_state in all_sensor_states.values():
            # only migrate instigators with the name "default_auto_materialize_sensor" and are
            # handled by the asset daemon
            if instigator_state.origin.instigator_name != "default_auto_materialize_sensor" and (
                instigator_state.sensor_instigator_data
                and instigator_state.sensor_instigator_data.sensor_type
                and instigator_state.sensor_instigator_data.sensor_type.is_handled_by_asset_daemon
            ):
                continue
            new_sensor_origin = instigator_state.origin._replace(
                instigator_name="default_automation_condition_sensor"
            )
            new_auto_materialize_state = InstigatorState(
                new_sensor_origin,
                InstigatorType.SENSOR,
                instigator_state.status,
                instigator_state.instigator_data,
            )
            new_sensor_selector_id = new_sensor_origin.get_selector().get_id()
            result[new_sensor_selector_id] = new_auto_materialize_state
            if all_sensor_states.get(new_sensor_selector_id):
                instance.update_instigator_state(new_auto_materialize_state)
            else:
                instance.add_instigator_state(new_auto_materialize_state)

        return result

    def _process_auto_materialize_tick_generator(
        self,
        workspace_process_context: IWorkspaceProcessContext,
        repository: Optional[RemoteRepository],
        sensor: Optional[RemoteSensor],
        debug_crash_flags: SingleInstigatorDebugCrashFlags,  # TODO No longer single instigator
        submit_threadpool_executor: Optional[ThreadPoolExecutor],
    ):
        evaluation_time = get_current_datetime()

        workspace = workspace_process_context.create_request_context()

        workspace_asset_graph = workspace.asset_graph

        instance: DagsterInstance = workspace_process_context.instance
        error_info = None

        if sensor:
            auto_materialize_instigator_state = check.not_none(
                instance.get_instigator_state(sensor.get_remote_origin_id(), sensor.selector_id)
            )
            if is_under_min_interval(auto_materialize_instigator_state, sensor):
                # check the since we might have been queued before processing
                return
            else:
                mark_sensor_state_for_tick(
                    instance, sensor, auto_materialize_instigator_state, evaluation_time
                )

        else:
            auto_materialize_instigator_state = None

        try:
            print_group_name = self._get_print_sensor_name(sensor)

            if sensor:
                selection = check.not_none(sensor.asset_selection)
                repository_origin = check.not_none(repository).get_remote_origin()
                # resolve the selection against just the assets in the sensor's repository
                repo_asset_graph = check.not_none(repository).asset_graph
                resolved_keys = selection.resolve(repo_asset_graph) | selection.resolve_checks(
                    repo_asset_graph
                )
                eligibility_graph = repo_asset_graph

                # Ensure that if there are two identical asset keys defined in different code
                # locations with automation conditions, only one of them actually launches runs
                eligible_keys = {
                    key
                    for key in resolved_keys
                    if (
                        workspace_asset_graph.get_repository_handle(key).get_remote_origin()
                        == repository_origin
                    )
                }
            else:
                eligible_keys = workspace_asset_graph.get_all_asset_keys()
                eligibility_graph = workspace_asset_graph

            auto_materialize_entity_keys = {
                target_key
                for target_key in eligible_keys
                if eligibility_graph.get(target_key).automation_condition is not None
            }
            num_target_entities = len(auto_materialize_entity_keys)

            auto_observe_asset_keys = {
                key
                for key in eligible_keys
                if isinstance(key, AssetKey)
                and eligibility_graph.get(key).auto_observe_interval_minutes is not None
            }
            num_auto_observe_assets = len(auto_observe_asset_keys)

            if not auto_materialize_entity_keys and not auto_observe_asset_keys:
                self._logger.debug(f"No assets/checks that require evaluation{print_group_name}")
                yield
                return

            self._logger.info(
                f"Checking {num_target_entities} assets/checks and"
                f" {num_auto_observe_assets} observable source"
                f" asset{'' if num_auto_observe_assets == 1 else 's'}{print_group_name}"
            )

            if sensor:
                stored_cursor = asset_daemon_cursor_from_instigator_serialized_cursor(
                    cast(
                        SensorInstigatorData,
                        check.not_none(auto_materialize_instigator_state).instigator_data,
                    ).cursor,
                    workspace_asset_graph,
                )

                instigator_origin_id = sensor.get_remote_origin().get_id()
                instigator_selector_id = sensor.get_remote_origin().get_selector().get_id()
                instigator_name = sensor.name
            else:
                stored_cursor = _get_pre_sensor_auto_materialize_cursor(
                    instance, workspace_asset_graph
                )
                instigator_origin_id = _PRE_SENSOR_AUTO_MATERIALIZE_ORIGIN_ID
                instigator_selector_id = _PRE_SENSOR_AUTO_MATERIALIZE_SELECTOR_ID
                instigator_name = _PRE_SENSOR_AUTO_MATERIALIZE_INSTIGATOR_NAME

            tick_retention_settings = instance.get_tick_retention_settings(
                InstigatorType.SENSOR if sensor else InstigatorType.AUTO_MATERIALIZE
            )

            ticks = instance.get_ticks(instigator_origin_id, instigator_selector_id, limit=1)
            latest_tick = ticks[0] if ticks else None

            max_retries = instance.auto_materialize_max_tick_retries

            # Determine if the most recent tick requires retrying
            retry_tick: Optional[InstigatorTick] = None
            override_evaluation_id: Optional[int] = None
            consecutive_failure_count: int = 0
            if latest_tick:
                can_resume = (
                    get_current_timestamp() - latest_tick.timestamp
                ) <= MAX_TIME_TO_RESUME_TICK_SECONDS

                if latest_tick.status in {TickStatus.FAILURE, TickStatus.STARTED}:
                    consecutive_failure_count = (
                        latest_tick.consecutive_failure_count or latest_tick.failure_count
                    )

                # the evaluation ids not matching indicates that the tick failed or crashed before
                # the cursor could be written, so no new runs could have been launched and it's
                # safe to re-evaluate things from scratch in a new tick without retrying anything
                previous_cursor_written = (
                    latest_tick.automation_condition_evaluation_id == stored_cursor.evaluation_id
                )

                if can_resume and not previous_cursor_written:
                    # if the tick failed before writing a cursor, we don't want to advance the
                    # evaluation id yet
                    override_evaluation_id = latest_tick.automation_condition_evaluation_id

                # If the previous tick matches the stored cursor's evaluation ID, check if it failed
                # or crashed partway through execution and needs to be resumed
                # Don't resume very old ticks though in case the daemon crashed for a long time and
                # then restarted
                if can_resume and previous_cursor_written:
                    if latest_tick.status == TickStatus.STARTED:
                        self._logger.warn(
                            f"Tick for evaluation {stored_cursor.evaluation_id}{print_group_name} was interrupted part-way through, resuming"
                        )
                        retry_tick = latest_tick
                    elif (
                        latest_tick.status == TickStatus.FAILURE
                        and latest_tick.tick_data.failure_count <= max_retries
                    ):
                        self._logger.info(
                            f"Retrying failed tick for evaluation {stored_cursor.evaluation_id}{print_group_name}"
                        )
                        retry_tick = instance.create_tick(
                            latest_tick.tick_data.with_status(
                                TickStatus.STARTED,
                                error=None,
                                timestamp=evaluation_time.timestamp(),
                                end_timestamp=None,
                            )._replace(
                                # make sure to override the evaluation id to stay on the previous value
                                auto_materialize_evaluation_id=latest_tick.automation_condition_evaluation_id
                            )
                        )
                    # otherwise, tick completed normally, no need to do anything
                else:
                    if latest_tick.status == TickStatus.STARTED:
                        # Old tick that won't be resumed - move it into a SKIPPED state so it isn't
                        # left dangling in STARTED
                        self._logger.warn(
                            f"Moving dangling STARTED tick from evaluation {latest_tick.automation_condition_evaluation_id}{print_group_name} into SKIPPED"
                        )
                        latest_tick = latest_tick.with_status(status=TickStatus.SKIPPED)
                        instance.update_tick(latest_tick)

            if retry_tick:
                tick = retry_tick
            else:
                tick = instance.create_tick(
                    TickData(
                        instigator_origin_id=instigator_origin_id,
                        instigator_name=instigator_name,
                        instigator_type=(
                            InstigatorType.SENSOR if sensor else InstigatorType.AUTO_MATERIALIZE
                        ),
                        status=TickStatus.STARTED,
                        timestamp=evaluation_time.timestamp(),
                        selector_id=instigator_selector_id,
                        consecutive_failure_count=consecutive_failure_count,
                        # we only set the auto_materialize_evaluation_id if it is not equal to the
                        # current tick id
                        auto_materialize_evaluation_id=override_evaluation_id,
                    )
                )

            with AutoMaterializeLaunchContext(
                tick,
                sensor,
                instance,
                self._logger,
                tick_retention_settings,
            ) as tick_context:
                yield from self._evaluate_auto_materialize_tick(
                    tick_context,
                    tick,
                    sensor,
                    workspace_process_context,
                    workspace_asset_graph,
                    auto_materialize_entity_keys,
                    stored_cursor,
                    auto_observe_asset_keys,
                    debug_crash_flags,
                    is_retry=(retry_tick is not None),
                    submit_threadpool_executor=submit_threadpool_executor,
                )
        except Exception:
            error_info = DaemonErrorCapture.process_exception(
                exc_info=sys.exc_info(),
                logger=self._logger,
                log_message="Automation condition daemon caught an error",
            )

        yield error_info

    _process_auto_materialize_tick = return_as_list(_process_auto_materialize_tick_generator)

    def _evaluate_auto_materialize_tick(
        self,
        tick_context: AutoMaterializeLaunchContext,
        tick: InstigatorTick,
        sensor: Optional[RemoteSensor],
        workspace_process_context: IWorkspaceProcessContext,
        asset_graph: RemoteAssetGraph,
        auto_materialize_entity_keys: set[EntityKey],
        stored_cursor: AssetDaemonCursor,
        auto_observe_asset_keys: set[AssetKey],
        debug_crash_flags: SingleInstigatorDebugCrashFlags,
        is_retry: bool,
        submit_threadpool_executor: Optional[ThreadPoolExecutor],
    ):
        evaluation_id = tick.automation_condition_evaluation_id

        instance = workspace_process_context.instance

        schedule_storage = check.not_none(instance.schedule_storage)

        if is_retry:
            # Unfinished or retried tick already generated evaluations and run requests and cursor, now
            # need to finish it
            run_requests = tick.tick_data.run_requests or []
            reserved_run_ids = tick.tick_data.reserved_run_ids or []

            if schedule_storage.supports_auto_materialize_asset_evaluations:
                evaluation_records = (
                    schedule_storage.get_auto_materialize_evaluations_for_evaluation_id(
                        evaluation_id
                    )
                )
                evaluations_by_key = {
                    evaluation_record.key: evaluation_record.get_evaluation_with_run_ids()
                    for evaluation_record in evaluation_records
                }
            else:
                evaluations_by_key = {}
        else:
            sensor_tags = {SENSOR_NAME_TAG: sensor.name, **sensor.run_tags} if sensor else {}

            # mold this into a shape AutomationTickEvaluationContext expects
            asset_selection = AssetSelection.keys(
                *{key for key in auto_materialize_entity_keys if isinstance(key, AssetKey)}
            ).without_checks() | AssetSelection.checks(
                *{key for key in auto_materialize_entity_keys if isinstance(key, AssetCheckKey)}
            )

            run_requests, new_cursor, evaluations = AutomationTickEvaluationContext(
                evaluation_id=evaluation_id,
                asset_graph=asset_graph,
                asset_selection=asset_selection,
                instance=instance,
                cursor=stored_cursor,
                materialize_run_tags={
                    **instance.auto_materialize_run_tags,
                    **DagsterRun.tags_for_tick_id(
                        str(tick.tick_id),
                    ),
                    **sensor_tags,
                },
                observe_run_tags={AUTO_OBSERVE_TAG: "true", **sensor_tags},
                emit_backfills=bool(
                    sensor
                    and sensor.metadata
                    and sensor.metadata.standard_metadata
                    and EMIT_BACKFILLS_METADATA_KEY in sensor.metadata.standard_metadata
                ),
                auto_observe_asset_keys=auto_observe_asset_keys,
                logger=self._logger,
            ).evaluate()

            check.invariant(new_cursor.evaluation_id == evaluation_id)

            check_for_debug_crash(debug_crash_flags, "EVALUATIONS_FINISHED")

            evaluations_by_key = {
                evaluation.key: evaluation.with_run_ids(set()) for evaluation in evaluations
            }

            # Write the asset evaluations without run IDs first
            if schedule_storage.supports_auto_materialize_asset_evaluations:
                schedule_storage.add_auto_materialize_asset_evaluations(
                    evaluation_id,
                    list(evaluations_by_key.values()),
                )
                check_for_debug_crash(debug_crash_flags, "ASSET_EVALUATIONS_ADDED")

            reserved_run_ids = [
                make_new_backfill_id() if rr.requires_backfill_daemon() else make_new_run_id()
                for rr in run_requests
            ]

            # Write out the in-progress tick data, which ensures that if the tick crashes or raises an exception, it will retry
            tick = tick_context.set_run_requests(
                run_requests=run_requests,
                reserved_run_ids=reserved_run_ids,
            )
            tick_context.write()
            check_for_debug_crash(debug_crash_flags, "RUN_REQUESTS_CREATED")

            # Write out the persistent cursor, which ensures that future ticks will move on once
            # they determine that nothing needs to be retried
            if sensor:
                state = instance.get_instigator_state(
                    sensor.get_remote_origin_id(), sensor.selector_id
                )
                instance.update_instigator_state(
                    check.not_none(state).with_data(
                        SensorInstigatorData(
                            last_tick_timestamp=tick.timestamp,
                            min_interval=sensor.min_interval_seconds,
                            cursor=asset_daemon_cursor_to_instigator_serialized_cursor(new_cursor),
                            sensor_type=sensor.sensor_type,
                        )
                    )
                )
            else:
                instance.daemon_cursor_storage.set_cursor_values(
                    {_PRE_SENSOR_AUTO_MATERIALIZE_CURSOR_KEY: serialize_value(new_cursor)}
                )

            check_for_debug_crash(debug_crash_flags, "CURSOR_UPDATED")

        print_group_name = self._get_print_sensor_name(sensor)

        self._logger.info(
            "Tick produced"
            f" {len(run_requests)} run{'s' if len(run_requests) != 1 else ''} and"
            f" {len(evaluations_by_key)} asset"
            f" evaluation{'s' if len(evaluations_by_key) != 1 else ''} for evaluation ID"
            f" {evaluation_id}{print_group_name}"
        )

        check.invariant(len(run_requests) == len(reserved_run_ids))
        yield from self._submit_run_requests_and_update_evaluations(
            instance=instance,
            tick_context=tick_context,
            workspace_process_context=workspace_process_context,
            evaluations_by_key=evaluations_by_key,
            evaluation_id=evaluation_id,
            run_requests=run_requests,
            reserved_run_ids=reserved_run_ids,
            debug_crash_flags=debug_crash_flags,
            submit_threadpool_executor=submit_threadpool_executor,
            remote_sensor=sensor,
        )

        if schedule_storage.supports_auto_materialize_asset_evaluations:
            schedule_storage.purge_asset_evaluations(
                before=(
                    get_current_datetime() - datetime.timedelta(days=EVALUATIONS_TTL_DAYS)
                ).timestamp(),
            )

        self._logger.info(f"Finished auto-materialization tick{print_group_name}")

    def _submit_run_request(
        self,
        i: int,
        instance: DagsterInstance,
        workspace_process_context: IWorkspaceProcessContext,
        evaluation_id: int,
        run_request: RunRequest,
        reserved_run_id: str,
        run_request_execution_data_cache: dict[JobSubsetSelector, RunRequestExecutionData],
        debug_crash_flags: SingleInstigatorDebugCrashFlags,
    ) -> tuple[str, AbstractSet[EntityKey]]:
        # check that the run_request requires the backfill daemon rather than if the setting is enabled to
        # account for the setting changing between tick retries
        if run_request.requires_backfill_daemon():
            asset_graph_subset = check.not_none(run_request.asset_graph_subset)
            if instance.get_backfill(reserved_run_id):
                self._logger.warn(
                    f"Run {reserved_run_id} already submitted on a previously interrupted tick, skipping"
                )
            else:
                instance.add_backfill(
                    PartitionBackfill.from_asset_graph_subset(
                        backfill_id=reserved_run_id,
                        dynamic_partitions_store=instance,
                        backfill_timestamp=get_current_timestamp(),
                        asset_graph_subset=asset_graph_subset,
                        tags={
                            **run_request.tags,
                            AUTO_MATERIALIZE_TAG: "true",
                            AUTOMATION_CONDITION_TAG: "true",
                            ASSET_EVALUATION_ID_TAG: str(evaluation_id),
                        },
                        title=f"Run for Declarative Automation evaluation ID {evaluation_id}",
                        description=None,
                    )
                )
            return reserved_run_id, check.not_none(asset_graph_subset.asset_keys)
        else:
            submitted_run = submit_asset_run(
                run_id=reserved_run_id,
                run_request=run_request._replace(
                    tags={
                        **run_request.tags,
                        AUTO_MATERIALIZE_TAG: "true",
                        AUTOMATION_CONDITION_TAG: "true",
                        ASSET_EVALUATION_ID_TAG: str(evaluation_id),
                    }
                ),
                run_request_index=i,
                instance=instance,
                workspace_process_context=workspace_process_context,
                run_request_execution_data_cache=run_request_execution_data_cache,
                debug_crash_flags=debug_crash_flags,
                logger=self._logger,
            )
            entity_keys = {
                *(run_request.asset_selection or []),
                *(run_request.asset_check_keys or []),
            }
            return submitted_run.run_id, entity_keys

    def _submit_run_requests_and_update_evaluations(
        self,
        instance: DagsterInstance,
        tick_context: AutoMaterializeLaunchContext,
        workspace_process_context: IWorkspaceProcessContext,
        evaluations_by_key: dict[EntityKey, AutomationConditionEvaluationWithRunIds],
        evaluation_id: int,
        run_requests: Sequence[RunRequest],
        reserved_run_ids: Sequence[str],
        debug_crash_flags: SingleInstigatorDebugCrashFlags,
        submit_threadpool_executor: Optional[ThreadPoolExecutor],
        remote_sensor: Optional[RemoteSensor],
    ):
        updated_evaluation_keys = set()
        run_request_execution_data_cache = {}
        check_after_runs_num = instance.get_tick_termination_check_interval()

        check.invariant(len(run_requests) == len(reserved_run_ids))
        to_submit = enumerate(tick_context.tick.reserved_run_ids_with_requests)

        def submit_run_request(
            run_id_with_run_request: tuple[int, tuple[str, RunRequest]],
        ) -> tuple[str, AbstractSet[EntityKey]]:
            i, (run_id, run_request) = run_id_with_run_request
            return self._submit_run_request(
                i=i,
                instance=instance,
                run_request=run_request,
                reserved_run_id=run_id,
                evaluation_id=evaluation_id,
                run_request_execution_data_cache=run_request_execution_data_cache,
                workspace_process_context=workspace_process_context,
                debug_crash_flags=debug_crash_flags,
            )

        if submit_threadpool_executor:
            gen_run_request_results = submit_threadpool_executor.map(submit_run_request, to_submit)
        else:
            gen_run_request_results = map(submit_run_request, to_submit)

        for i, (submitted_run_id, entity_keys) in enumerate(gen_run_request_results):
            # heartbeat after each submitted run
            yield

            tick_context.add_run_info(run_id=submitted_run_id)

            # write the submitted run ID to any evaluations
            for entity_key in entity_keys:
                # asset keys for observation runs don't have evaluations
                if entity_key in evaluations_by_key:
                    evaluation = evaluations_by_key[entity_key]
                    evaluations_by_key[entity_key] = dataclasses.replace(
                        evaluation, run_ids=evaluation.run_ids | {submitted_run_id}
                    )
                    updated_evaluation_keys.add(entity_key)

            # check if the sensor is still enabled:
            if check_after_runs_num is not None and i % check_after_runs_num == 0:
                if not self._sensor_is_enabled(instance, remote_sensor):
                    # The user has manually stopped the sensor mid-iteration. In this case we assume
                    # the user has a good reason for stopping the sensor (e.g. the sensor is submitting
                    # many unintentional runs) so we stop submitting runs and will mark the tick as
                    # skipped so that when the sensor is turned back on we don't detect this tick as incomplete
                    # and try to submit the same runs again.
                    self._logger.info(
                        "Sensor has been manually stopped while submitted runs. No more runs will be submitted."
                    )
                    tick_context.set_user_interrupted(True)
                    break

        evaluations_to_update = [
            evaluations_by_key[asset_key] for asset_key in updated_evaluation_keys
        ]
        if evaluations_to_update:
            schedule_storage = check.not_none(instance.schedule_storage)
            schedule_storage.add_auto_materialize_asset_evaluations(
                evaluation_id, evaluations_to_update
            )

        check_for_debug_crash(debug_crash_flags, "RUN_IDS_ADDED_TO_EVALUATIONS")

        if tick_context.tick.tick_data.user_interrupted:
            # mark as skipped so that we don't request any remaining runs when the sensor is started again
            tick_context.update_state(TickStatus.SKIPPED)
            tick_context.set_skip_reason("Sensor manually stopped mid-iteration.")
        else:
            tick_context.update_state(
                TickStatus.SUCCESS if len(run_requests) > 0 else TickStatus.SKIPPED,
            )

    def _sensor_is_enabled(self, instance: DagsterInstance, remote_sensor: Optional[RemoteSensor]):
        use_auto_materialize_sensors = instance.auto_materialize_use_sensors
        if (not use_auto_materialize_sensors) and get_auto_materialize_paused(instance):
            return False
        if use_auto_materialize_sensors and remote_sensor:
            instigator_state = instance.get_instigator_state(
                remote_sensor.get_remote_origin_id(), remote_sensor.selector_id
            )
            if instigator_state and not instigator_state.is_running:
                return False

        return True



from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Optional, TypeVar, cast

from dagster._core.definitions.cacheable_assets import AssetsDefinitionCacheableData
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.errors import DagsterInvariantViolationError
from dagster._serdes.serdes import PackableValue, deserialize_value, serialize_value

if TYPE_CHECKING:
    from dagster._core.definitions.repository_definition import RepositoryLoadData


class DefinitionsLoadType(Enum):
    """An enum that defines the different scenarios in which a Definitions object may be
    instantiated.

    When booting a Dagster process that contains definitions, one of two things are happening.

    1. `INITIALIZATION`: You are defining a set of definitions for Dagster to ingest, effectively updating Dagster's
        understanding of the world at a point in time.
    2. `RECONSTRUCTION`: You are reconstructing a set of definitions that were previously defined.

    This distinction is important when the source of truth for a Dagster definition is state
    (for example in a hosted ingestion or BI tool accessable via an API). The state should
    only be accessed when defining definitions, whereas when reconstructing definitions that
    state should be retrieved from Dagster's metastore.
    """

    INITIALIZATION = "INITIALIZATION"
    RECONSTRUCTION = "RECONSTRUCTION"


class DefinitionsLoadContext:
    """Holds data that's made available to Definitions-loading code when a DefinitionsLoader is
    invoked.
    User construction of this object is not supported.
    """

    _instance: ClassVar[Optional["DefinitionsLoadContext"]] = None

    def __init__(
        self,
        load_type: DefinitionsLoadType,
        repository_load_data: Optional["RepositoryLoadData"] = None,
    ):
        self._load_type = load_type
        self._repository_load_data = repository_load_data
        self._pending_reconstruction_metadata = {}

    @classmethod
    def get(cls) -> "DefinitionsLoadContext":
        """Get the current DefinitionsLoadContext. If it has not been set, the
        context is assumed to be initialization.
        """
        return DefinitionsLoadContext._instance or cls(load_type=DefinitionsLoadType.INITIALIZATION)

    @classmethod
    def set(cls, instance: "DefinitionsLoadContext") -> None:
        """Get the current DefinitionsLoadContext."""
        cls._instance = instance

    @property
    def load_type(self) -> DefinitionsLoadType:
        """DefinitionsLoadType: Classifier for scenario in which Definitions are being loaded."""
        return self._load_type

    def add_to_pending_reconstruction_metadata(self, key: str, metadata: Any) -> None:
        self._pending_reconstruction_metadata[key] = metadata

    def get_pending_reconstruction_metadata(self) -> Mapping[str, Any]:
        return self._pending_reconstruction_metadata

    @property
    def cacheable_asset_data(self) -> Mapping[str, Sequence[AssetsDefinitionCacheableData]]:
        """Sequence[AssetDefinitionCacheableData]: A sequence of cacheable asset data created
        during code location initialization. Accessing this during initialization will raise an
        error.
        """
        if self._load_type == DefinitionsLoadType.INITIALIZATION:
            raise DagsterInvariantViolationError(
                "Attempted to access cacheable asset data during code location initialization."
                " Cacheable asset data is only available during reconstruction of a code location."
            )
        return self._repository_load_data.cacheable_asset_data if self._repository_load_data else {}

    @cached_property
    def reconstruction_metadata(self) -> Mapping[str, Any]:
        """Mapping[str, Any]: A dictionary containing metadata from the returned Definitions object
        at initial code location initialization. Accessing this during initialization will raise an
        error.
        """
        if self._load_type == DefinitionsLoadType.INITIALIZATION:
            raise DagsterInvariantViolationError(
                "Attempted to access code location metadata during code location initialization."
                " Code location metadata is only available during reconstruction of a code location."
            )
        # Expose the wrapped metadata values so that users access exactly what they put in.
        return (
            {k: v.data for k, v in self._repository_load_data.reconstruction_metadata.items()}
            if self._repository_load_data
            else {}
        )


TState = TypeVar("TState", bound=PackableValue)


class StateBackedDefinitionsLoader(ABC, Generic[TState]):
    """A class for building state-backed definitions. In a structured way. It
    handles manipulation of the DefinitionsLoadContext on the user's behalf.

    The goal here is so that an unreliable backing source is fetched only
    at code server load time, which defines the code location. When, for example,
    a run worker is launched, the backing source is not queryed again, and
    only `defs_from_state` is called.

    Current meant for internal usage only hence TState must be a marked as whitelist_for_serdes.

    Args:
        defs_key (str): The unique key for the definitions. Must be unique per code location.

    Examples:

    .. code-block:: python

        @whitelist_for_serdes
        @record
        class ExampleDefState:
            a_string: str

        class ExampleStateBackedDefinitionsLoader(StateBackedDefinitionsLoader[ExampleDefState]):
            def fetch_state(self) -> ExampleDefState:
                # Fetch from potentially unreliable API (e.g. Rest API)
                return ExampleDefState(a_string="foo")

            def defs_from_state(self, state: ExampleDefState) -> Definitions:
                # Construct or reconstruct the Definitions from the previously
                # fetched state.
                return Definitions([AssetSpec(key=state.a_string)])
    """

    @property
    @abstractmethod
    def defs_key(self) -> str:
        """The unique key for the definitions. Must be unique per code location."""
        ...

    @abstractmethod
    def fetch_state(self) -> TState:
        """Subclasses must implement this method. This is where the integration runs
        code that fetches the backing state from the source of truth for the definitions.
        This is only called when the code location is initializing, for example on
        code server load, or when loading via dagster dev.
        """
        ...

    @abstractmethod
    def defs_from_state(self, state: TState) -> Definitions:
        """Subclasses must implement this method. It is invoked whenever the code location
        is loading, whether it be initializaton or reconstruction. In the case of
        intialization, it takes the result of fetch_backing state that just happened.
        When reconstructing, it takes the state that was previously fetched and attached
        as metadata.

        This method also takes responsibility for attaching the state to the definitions
        on its metadata, with the key passed in as defs_key.
        """
        ...

    def get_or_fetch_state(self) -> TState:
        context = DefinitionsLoadContext.get()
        state = (
            cast(TState, deserialize_value(context.reconstruction_metadata[self.defs_key]))
            if (
                context.load_type == DefinitionsLoadType.RECONSTRUCTION
                and self.defs_key in context.reconstruction_metadata
            )
            else self.fetch_state()
        )
        context.add_to_pending_reconstruction_metadata(self.defs_key, serialize_value(state))

        return state

    def build_defs(self) -> Definitions:
        state = self.get_or_fetch_state()

        return self.defs_from_state(state).with_reconstruction_metadata(
            {self.defs_key: serialize_value(state)}
        )



from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, Optional, Union  # noqa: UP035

from dagster import _check as check
from dagster._config.config_schema import UserConfigSchema
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_key import EntityKey
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionEvaluationWithRunIds,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.event_api import EventHandlerFn
from dagster._core.storage.asset_check_execution_record import (
    AssetCheckExecutionRecord,
    AssetCheckExecutionRecordStatus,
)
from dagster._core.storage.base_storage import DagsterStorage
from dagster._core.storage.event_log.base import (
    AssetCheckSummaryRecord,
    AssetRecord,
    EventLogConnection,
    EventLogRecord,
    EventLogStorage,
    EventRecordsFilter,
    EventRecordsResult,
    PlannedMaterializationInfo,
    PoolLimit,
)
from dagster._core.storage.runs.base import RunStorage
from dagster._core.storage.schedules.base import ScheduleStorage
from dagster._core.storage.sql import AlembicVersion
from dagster._serdes import ConfigurableClass, ConfigurableClassData
from dagster._utils import PrintFn
from dagster._utils.concurrency import ConcurrencyClaimStatus, ConcurrencyKeyInfo

if TYPE_CHECKING:
    from dagster._core.definitions.asset_check_spec import AssetCheckKey
    from dagster._core.definitions.run_request import InstigatorType
    from dagster._core.event_api import AssetRecordsFilter, RunStatusChangeRecordsFilter
    from dagster._core.events import DagsterEvent, DagsterEventType
    from dagster._core.events.log import EventLogEntry
    from dagster._core.execution.backfill import (
        BulkActionsFilter,
        BulkActionStatus,
        PartitionBackfill,
    )
    from dagster._core.execution.stats import RunStepKeyStatsSnapshot
    from dagster._core.instance import DagsterInstance
    from dagster._core.remote_representation.origin import RemoteJobOrigin
    from dagster._core.scheduler.instigation import (
        AutoMaterializeAssetEvaluationRecord,
        InstigatorState,
        InstigatorStatus,
        InstigatorTick,
        TickData,
        TickStatus,
    )
    from dagster._core.snap.execution_plan_snapshot import ExecutionPlanSnapshot
    from dagster._core.snap.job_snapshot import JobSnap
    from dagster._core.storage.dagster_run import (
        DagsterRun,
        DagsterRunStatsSnapshot,
        JobBucket,
        RunPartitionData,
        RunRecord,
        RunsFilter,
        TagBucket,
    )
    from dagster._core.storage.partition_status_cache import AssetStatusCacheValue
    from dagster._daemon.types import DaemonHeartbeat


class CompositeStorage(DagsterStorage, ConfigurableClass):
    """Utiltity class for combining the individually configured run, event_log, schedule storages
    into the single dagster storage.
    """

    def __init__(
        self,
        run_storage: RunStorage,
        event_log_storage: EventLogStorage,
        schedule_storage: ScheduleStorage,
        inst_data: Optional[ConfigurableClassData] = None,
    ):
        self._run_storage = run_storage
        self._event_log_storage = check.inst_param(
            event_log_storage, "event_log_storage", EventLogStorage
        )
        self._schedule_storage = check.inst_param(
            schedule_storage, "schedule_storage", ScheduleStorage
        )
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        super().__init__()

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {
            "run_storage": {
                "module_name": str,
                "class_name": str,
                "config_yaml": str,
            },
            "event_log_storage": {
                "module_name": str,
                "class_name": str,
                "config_yaml": str,
            },
            "schedule_storage": {
                "module_name": str,
                "class_name": str,
                "config_yaml": str,
            },
        }

    @classmethod
    def from_config_value(
        cls,
        inst_data: Optional[ConfigurableClassData],
        config_value: Mapping[str, Mapping[str, str]],
    ) -> "CompositeStorage":
        run_storage_config = config_value["run_storage"]
        run_storage = ConfigurableClassData(
            module_name=run_storage_config["module_name"],
            class_name=run_storage_config["class_name"],
            config_yaml=run_storage_config["config_yaml"],
        ).rehydrate(as_type=RunStorage)
        event_log_storage_config = config_value["event_log_storage"]
        event_log_storage = ConfigurableClassData(
            module_name=event_log_storage_config["module_name"],
            class_name=event_log_storage_config["class_name"],
            config_yaml=event_log_storage_config["config_yaml"],
        ).rehydrate(as_type=EventLogStorage)
        schedule_storage_config = config_value["schedule_storage"]
        schedule_storage = ConfigurableClassData(
            module_name=schedule_storage_config["module_name"],
            class_name=schedule_storage_config["class_name"],
            config_yaml=schedule_storage_config["config_yaml"],
        ).rehydrate(as_type=ScheduleStorage)
        return CompositeStorage(
            run_storage, event_log_storage, schedule_storage, inst_data=inst_data
        )

    @property
    def event_log_storage(self) -> EventLogStorage:
        return self._event_log_storage

    @property
    def run_storage(self) -> RunStorage:
        return self._run_storage

    @property
    def schedule_storage(self) -> ScheduleStorage:
        return self._schedule_storage


class LegacyRunStorage(RunStorage, ConfigurableClass):
    def __init__(self, storage: DagsterStorage, inst_data: Optional[ConfigurableClassData] = None):
        self._storage = check.inst_param(storage, "storage", DagsterStorage)
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        super().__init__()

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @property
    def _instance(self) -> Optional["DagsterInstance"]:
        return self._storage._instance  # noqa: SLF001

    def register_instance(self, instance: "DagsterInstance") -> None:
        if not self._storage.has_instance:
            self._storage.register_instance(instance)

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {
            "module_name": str,
            "class_name": str,
            "config_yaml": str,
        }

    @classmethod
    def from_config_value(
        cls, inst_data: Optional[ConfigurableClassData], config_value: Mapping[str, str]
    ) -> "LegacyRunStorage":
        storage = ConfigurableClassData(
            module_name=config_value["module_name"],
            class_name=config_value["class_name"],
            config_yaml=config_value["config_yaml"],
        ).rehydrate(as_type=DagsterStorage)
        return LegacyRunStorage(storage, inst_data=inst_data)

    def add_run(self, dagster_run: "DagsterRun") -> "DagsterRun":
        return self._storage.run_storage.add_run(dagster_run)

    def handle_run_event(self, run_id: str, event: "DagsterEvent") -> None:
        return self._storage.run_storage.handle_run_event(run_id, event)

    def get_runs(
        self,
        filters: Optional["RunsFilter"] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        bucket_by: Optional[Union["JobBucket", "TagBucket"]] = None,
        ascending: bool = False,
    ) -> Iterable["DagsterRun"]:
        return self._storage.run_storage.get_runs(filters, cursor, limit, bucket_by, ascending)

    def get_run_ids(
        self,
        filters: Optional["RunsFilter"] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Iterable[str]:
        return self._storage.run_storage.get_run_ids(filters, cursor=cursor, limit=limit)

    def get_runs_count(self, filters: Optional["RunsFilter"] = None) -> int:
        return self._storage.run_storage.get_runs_count(filters)

    def get_run_group(self, run_id: str) -> Optional[tuple[str, Iterable["DagsterRun"]]]:
        return self._storage.run_storage.get_run_group(run_id)

    def get_run_records(
        self,
        filters: Optional["RunsFilter"] = None,
        limit: Optional[int] = None,
        order_by: Optional[str] = None,
        ascending: bool = False,
        cursor: Optional[str] = None,
        bucket_by: Optional[Union["JobBucket", "TagBucket"]] = None,
    ) -> Sequence["RunRecord"]:
        return self._storage.run_storage.get_run_records(
            filters, limit, order_by, ascending, cursor, bucket_by
        )

    def get_run_tags(
        self,
        tag_keys: Sequence[str],
        value_prefix: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Sequence[tuple[str, set[str]]]:
        return self._storage.run_storage.get_run_tags(tag_keys, value_prefix, limit)

    def get_run_tag_keys(self) -> Sequence[str]:
        return self._storage.run_storage.get_run_tag_keys()

    def add_run_tags(self, run_id: str, new_tags: Mapping[str, str]):
        return self._storage.run_storage.add_run_tags(run_id, new_tags)

    def has_run(self, run_id: str) -> bool:
        return self._storage.run_storage.has_run(run_id)

    def add_snapshot(
        self,
        snapshot: Union["JobSnap", "ExecutionPlanSnapshot"],
    ) -> None:
        return self._storage.run_storage.add_snapshot(snapshot)

    def has_snapshot(self, snapshot_id: str) -> bool:
        return self._storage.run_storage.has_snapshot(snapshot_id)

    def has_job_snapshot(self, job_snapshot_id: str) -> bool:
        return self._storage.run_storage.has_job_snapshot(job_snapshot_id)

    def add_job_snapshot(self, job_snapshot: "JobSnap") -> str:
        return self._storage.run_storage.add_job_snapshot(job_snapshot)

    def get_job_snapshot(self, job_snapshot_id: str) -> "JobSnap":
        return self._storage.run_storage.get_job_snapshot(job_snapshot_id)

    def has_execution_plan_snapshot(self, execution_plan_snapshot_id: str) -> bool:
        return self._storage.run_storage.has_execution_plan_snapshot(execution_plan_snapshot_id)

    def add_execution_plan_snapshot(self, execution_plan_snapshot: "ExecutionPlanSnapshot") -> str:
        return self._storage.run_storage.add_execution_plan_snapshot(execution_plan_snapshot)

    def get_execution_plan_snapshot(
        self, execution_plan_snapshot_id: str
    ) -> "ExecutionPlanSnapshot":
        return self._storage.run_storage.get_execution_plan_snapshot(execution_plan_snapshot_id)

    def wipe(self) -> None:
        return self._storage.run_storage.wipe()

    def delete_run(self, run_id: str) -> None:
        return self._storage.run_storage.delete_run(run_id)

    def migrate(self, print_fn: Optional[PrintFn] = None, force_rebuild_all: bool = False) -> None:
        return self._storage.run_storage.migrate(print_fn, force_rebuild_all)

    def optimize(self, print_fn: Optional[PrintFn] = None, force_rebuild_all: bool = False) -> None:
        return self._storage.run_storage.optimize(print_fn, force_rebuild_all)

    def dispose(self) -> None:
        return self._storage.run_storage.dispose()

    def optimize_for_webserver(
        self, statement_timeout: int, pool_recycle: int, max_overflow: int
    ) -> None:
        return self._storage.run_storage.optimize_for_webserver(
            statement_timeout, pool_recycle, max_overflow
        )

    def add_daemon_heartbeat(self, daemon_heartbeat: "DaemonHeartbeat") -> None:
        return self._storage.run_storage.add_daemon_heartbeat(daemon_heartbeat)

    def get_daemon_heartbeats(self) -> Mapping[str, "DaemonHeartbeat"]:
        return self._storage.run_storage.get_daemon_heartbeats()

    def wipe_daemon_heartbeats(self) -> None:
        return self._storage.run_storage.wipe_daemon_heartbeats()

    def get_backfills(
        self,
        filters: Optional["BulkActionsFilter"] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        status: Optional["BulkActionStatus"] = None,
    ) -> Sequence["PartitionBackfill"]:
        return self._storage.run_storage.get_backfills(
            cursor=cursor, limit=limit, filters=filters, status=status
        )

    def get_backfills_count(self, filters: Optional["BulkActionsFilter"] = None) -> int:
        return self._storage.run_storage.get_backfills_count(filters=filters)

    def get_backfill(self, backfill_id: str) -> Optional["PartitionBackfill"]:
        return self._storage.run_storage.get_backfill(backfill_id)

    def add_backfill(self, partition_backfill: "PartitionBackfill") -> None:
        return self._storage.run_storage.add_backfill(partition_backfill)

    def update_backfill(self, partition_backfill: "PartitionBackfill") -> None:
        return self._storage.run_storage.update_backfill(partition_backfill)

    def get_run_partition_data(self, runs_filter: "RunsFilter") -> Sequence["RunPartitionData"]:
        return self._storage.run_storage.get_run_partition_data(runs_filter)

    def get_cursor_values(self, keys: set[str]) -> Mapping[str, str]:
        return self._storage.run_storage.get_cursor_values(keys)

    def set_cursor_values(self, pairs: Mapping[str, str]) -> None:
        return self._storage.run_storage.set_cursor_values(pairs)

    def replace_job_origin(self, run: "DagsterRun", job_origin: "RemoteJobOrigin") -> None:
        return self._storage.run_storage.replace_job_origin(run, job_origin)

    def alembic_version(self) -> Optional[AlembicVersion]:
        return self._storage.run_storage.alembic_version()


class LegacyEventLogStorage(EventLogStorage, ConfigurableClass):
    def __init__(self, storage: DagsterStorage, inst_data: Optional[ConfigurableClassData] = None):
        self._storage = check.inst_param(storage, "storage", DagsterStorage)
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        super().__init__()

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {
            "module_name": str,
            "class_name": str,
            "config_yaml": str,
        }

    @classmethod
    def from_config_value(
        cls, inst_data: Optional[ConfigurableClassData], config_value: Mapping[str, str]
    ) -> "LegacyEventLogStorage":
        storage = ConfigurableClassData(
            module_name=config_value["module_name"],
            class_name=config_value["class_name"],
            config_yaml=config_value["config_yaml"],
        ).rehydrate(as_type=DagsterStorage)
        # Type checker says LegacyEventStorage is abstract and can't be instantiated. Not sure whether
        # type check is wrong, or is unused code path.
        return LegacyEventLogStorage(storage, inst_data=inst_data)

    @property
    def _instance(self) -> Optional["DagsterInstance"]:
        return self._storage._instance  # noqa: SLF001

    def index_connection(self):
        return self._storage.event_log_storage.index_connection()  # pyright: ignore[reportAttributeAccessIssue]

    def register_instance(self, instance: "DagsterInstance") -> None:
        if not self._storage.has_instance:
            self._storage.register_instance(instance)

    def get_logs_for_run(
        self,
        run_id: str,
        cursor: Optional[Union[str, int]] = None,
        of_type: Optional[Union["DagsterEventType", set["DagsterEventType"]]] = None,
        limit: Optional[int] = None,
        ascending: bool = True,
    ) -> Iterable["EventLogEntry"]:
        return self._storage.event_log_storage.get_logs_for_run(
            run_id, cursor, of_type, limit, ascending
        )

    def get_logs_for_all_runs_by_log_id(
        self,
        after_cursor: int = -1,
        dagster_event_type: Optional[Union["DagsterEventType", set["DagsterEventType"]]] = None,
        limit: Optional[int] = None,
    ) -> Mapping[int, "EventLogEntry"]:
        return self._storage.event_log_storage.get_logs_for_all_runs_by_log_id(
            after_cursor=after_cursor, dagster_event_type=dagster_event_type, limit=limit
        )

    def get_maximum_record_id(self) -> Optional[int]:
        return self._storage.event_log_storage.get_maximum_record_id()

    def get_stats_for_run(self, run_id: str) -> "DagsterRunStatsSnapshot":
        return self._storage.event_log_storage.get_stats_for_run(run_id)

    def get_step_stats_for_run(
        self, run_id: str, step_keys: Optional[Sequence[str]] = None
    ) -> Sequence["RunStepKeyStatsSnapshot"]:
        return self._storage.event_log_storage.get_step_stats_for_run(run_id, step_keys)

    def store_event(self, event: "EventLogEntry") -> None:
        return self._storage.event_log_storage.store_event(event)

    def delete_events(self, run_id: str) -> None:
        return self._storage.event_log_storage.delete_events(run_id)

    def upgrade(self) -> None:
        return self._storage.event_log_storage.upgrade()

    def reindex_events(self, print_fn: Optional[PrintFn] = None, force: bool = False) -> None:
        return self._storage.event_log_storage.reindex_events(print_fn, force)

    def reindex_assets(self, print_fn: Optional[PrintFn] = None, force: bool = False) -> None:
        return self._storage.event_log_storage.reindex_assets(print_fn, force)

    def wipe(self) -> None:
        return self._storage.event_log_storage.wipe()

    def watch(self, run_id: str, cursor: str, callback: EventHandlerFn) -> None:
        return self._storage.event_log_storage.watch(run_id, cursor, callback)

    def end_watch(self, run_id: str, handler: EventHandlerFn) -> None:
        return self._storage.event_log_storage.end_watch(run_id, handler)

    @property
    def is_persistent(self) -> bool:
        return self._storage.event_log_storage.is_persistent

    def dispose(self) -> None:
        return self._storage.event_log_storage.dispose()

    def optimize_for_webserver(
        self, statement_timeout: int, pool_recycle: int, max_overflow: int
    ) -> None:
        return self._storage.event_log_storage.optimize_for_webserver(
            statement_timeout,
            pool_recycle,
            max_overflow,
        )

    def get_event_records(
        self,
        event_records_filter: Optional[EventRecordsFilter] = None,
        limit: Optional[int] = None,
        ascending: bool = False,
    ) -> Iterable[EventLogRecord]:
        # type ignored because `get_event_records` does not accept None. Unclear which type
        # annotation is wrong.
        return self._storage.event_log_storage.get_event_records(
            event_records_filter,  # type: ignore
            limit,
            ascending,
        )

    def fetch_materializations(
        self,
        filters: Union[AssetKey, "AssetRecordsFilter"],
        limit: int,
        cursor: Optional[str] = None,
        ascending: bool = False,
    ) -> EventRecordsResult:
        return self._storage.event_log_storage.fetch_materializations(
            filters, limit, cursor, ascending
        )

    def fetch_observations(
        self,
        filters: Union[AssetKey, "AssetRecordsFilter"],
        limit: int,
        cursor: Optional[str] = None,
        ascending: bool = False,
    ) -> EventRecordsResult:
        return self._storage.event_log_storage.fetch_observations(filters, limit, cursor, ascending)

    def fetch_run_status_changes(
        self,
        filters: Union["DagsterEventType", "RunStatusChangeRecordsFilter"],
        limit: int,
        cursor: Optional[str] = None,
        ascending: bool = False,
    ) -> EventRecordsResult:
        return self._storage.event_log_storage.fetch_run_status_changes(
            filters, limit, cursor, ascending
        )

    def get_latest_planned_materialization_info(
        self,
        asset_key: AssetKey,
        partition: Optional[str] = None,
    ) -> Optional[PlannedMaterializationInfo]:
        return self._storage.event_log_storage.get_latest_planned_materialization_info(
            asset_key, partition
        )

    def get_updated_data_version_partitions(
        self, asset_key: AssetKey, partitions: Iterable[str], since_storage_id: int
    ) -> set[str]:
        return self._storage.event_log_storage.get_updated_data_version_partitions(
            asset_key, partitions, since_storage_id
        )

    def get_asset_records(
        self, asset_keys: Optional[Sequence["AssetKey"]] = None
    ) -> Iterable[AssetRecord]:
        return self._storage.event_log_storage.get_asset_records(asset_keys)

    def get_asset_check_summary_records(
        self, asset_check_keys: Sequence["AssetCheckKey"]
    ) -> Mapping["AssetCheckKey", AssetCheckSummaryRecord]:
        return self._storage.event_log_storage.get_asset_check_summary_records(asset_check_keys)

    def has_asset_key(self, asset_key: "AssetKey") -> bool:
        return self._storage.event_log_storage.has_asset_key(asset_key)

    def all_asset_keys(self) -> Iterable["AssetKey"]:
        return self._storage.event_log_storage.all_asset_keys()

    def get_asset_keys(
        self,
        prefix: Optional[Sequence[str]] = None,
        limit: Optional[int] = None,
        cursor: Optional[str] = None,
    ) -> Iterable["AssetKey"]:
        return self._storage.event_log_storage.get_asset_keys(prefix, limit, cursor)

    def get_latest_materialization_events(
        self, asset_keys: Iterable["AssetKey"]
    ) -> Mapping["AssetKey", Optional["EventLogEntry"]]:
        return self._storage.event_log_storage.get_latest_materialization_events(asset_keys)

    def wipe_asset(self, asset_key: "AssetKey") -> None:
        return self._storage.event_log_storage.wipe_asset(asset_key)

    def wipe_asset_partitions(self, asset_key: AssetKey, partition_keys: Sequence[str]) -> None:
        """Remove asset index history from event log for given asset partitions."""
        raise NotImplementedError(
            "Partitioned asset wipe is not supported yet for this event log storage."
        )

    def get_materialized_partitions(
        self,
        asset_key: AssetKey,
        before_cursor: Optional[int] = None,
        after_cursor: Optional[int] = None,
    ) -> set[str]:
        return self._storage.event_log_storage.get_materialized_partitions(
            asset_key, before_cursor, after_cursor
        )

    def get_latest_storage_id_by_partition(
        self,
        asset_key: "AssetKey",
        event_type: "DagsterEventType",
        partitions: Optional[set[str]] = None,
    ) -> Mapping[str, int]:
        return self._storage.event_log_storage.get_latest_storage_id_by_partition(
            asset_key, event_type, partitions
        )

    def get_latest_tags_by_partition(
        self,
        asset_key: "AssetKey",
        event_type: "DagsterEventType",
        tag_keys: Sequence[str],
        asset_partitions: Optional[Sequence[str]] = None,
        before_cursor: Optional[int] = None,
        after_cursor: Optional[int] = None,
    ) -> Mapping[str, Mapping[str, str]]:
        return self._storage.event_log_storage.get_latest_tags_by_partition(
            asset_key, event_type, tag_keys, asset_partitions, before_cursor, after_cursor
        )

    def get_latest_asset_partition_materialization_attempts_without_materializations(
        self, asset_key: "AssetKey", after_storage_id: Optional[int] = None
    ) -> Mapping[str, tuple[str, int]]:
        return self._storage.event_log_storage.get_latest_asset_partition_materialization_attempts_without_materializations(
            asset_key, after_storage_id
        )

    def get_dynamic_partitions(self, partitions_def_name: str) -> Sequence[str]:
        return self._storage.event_log_storage.get_dynamic_partitions(partitions_def_name)

    def has_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> bool:
        return self._storage.event_log_storage.has_dynamic_partition(
            partitions_def_name, partition_key
        )

    def add_dynamic_partitions(
        self, partitions_def_name: str, partition_keys: Sequence[str]
    ) -> None:
        return self._storage.event_log_storage.add_dynamic_partitions(
            partitions_def_name, partition_keys
        )

    def delete_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> None:
        return self._storage.event_log_storage.delete_dynamic_partition(
            partitions_def_name, partition_key
        )

    def get_event_tags_for_asset(
        self,
        asset_key: "AssetKey",
        filter_tags: Optional[Mapping[str, str]] = None,
        filter_event_id: Optional[int] = None,
    ) -> Sequence[Mapping[str, str]]:
        return self._storage.event_log_storage.get_event_tags_for_asset(
            asset_key, filter_tags, filter_event_id
        )

    def can_read_asset_status_cache(self) -> bool:
        return self._storage.event_log_storage.can_read_asset_status_cache()

    def can_write_asset_status_cache(self) -> bool:
        return self._storage.event_log_storage.can_write_asset_status_cache()

    def wipe_asset_cached_status(self, asset_key: "AssetKey") -> None:
        return self._storage.event_log_storage.wipe_asset_cached_status(asset_key)

    def update_asset_cached_status_data(
        self, asset_key: "AssetKey", cache_values: "AssetStatusCacheValue"
    ) -> None:
        self._storage.event_log_storage.update_asset_cached_status_data(
            asset_key=asset_key, cache_values=cache_values
        )

    def get_records_for_run(
        self,
        run_id: str,
        cursor: Optional[str] = None,
        of_type: Optional[Union["DagsterEventType", set["DagsterEventType"]]] = None,
        limit: Optional[int] = None,
        ascending: bool = True,
    ) -> EventLogConnection:
        return self._storage.event_log_storage.get_records_for_run(
            run_id, cursor, of_type, limit, ascending
        )

    def initialize_concurrency_limit_to_default(self, concurrency_key: str) -> bool:
        return self._storage.event_log_storage.initialize_concurrency_limit_to_default(
            concurrency_key
        )

    def set_concurrency_slots(self, concurrency_key: str, num: int) -> None:
        return self._storage.event_log_storage.set_concurrency_slots(concurrency_key, num)

    def delete_concurrency_limit(self, concurrency_key: str) -> None:
        return self._storage.event_log_storage.delete_concurrency_limit(concurrency_key)

    def get_concurrency_keys(self) -> set[str]:
        return self._storage.event_log_storage.get_concurrency_keys()

    def get_pool_limits(self) -> Sequence[PoolLimit]:
        return self._storage.event_log_storage.get_pool_limits()

    def get_concurrency_info(self, concurrency_key: str) -> ConcurrencyKeyInfo:
        return self._storage.event_log_storage.get_concurrency_info(concurrency_key)

    def claim_concurrency_slot(
        self, concurrency_key: str, run_id: str, step_key: str, priority: Optional[int] = None
    ) -> ConcurrencyClaimStatus:
        return self._storage.event_log_storage.claim_concurrency_slot(
            concurrency_key, run_id, step_key, priority
        )

    def check_concurrency_claim(self, concurrency_key: str, run_id: str, step_key: str):
        return self._storage.event_log_storage.check_concurrency_claim(
            concurrency_key, run_id, step_key
        )

    def get_concurrency_run_ids(self) -> set[str]:
        return self._storage.event_log_storage.get_concurrency_run_ids()

    def free_concurrency_slots_for_run(self, run_id: str) -> None:
        return self._storage.event_log_storage.free_concurrency_slots_for_run(run_id)

    def free_concurrency_slot_for_step(self, run_id: str, step_key: str) -> None:
        return self._storage.event_log_storage.free_concurrency_slot_for_step(run_id, step_key)

    def get_asset_check_execution_history(
        self,
        check_key: "AssetCheckKey",
        limit: int,
        cursor: Optional[int] = None,
        status: Optional[AbstractSet[AssetCheckExecutionRecordStatus]] = None,
    ) -> Sequence[AssetCheckExecutionRecord]:
        return self._storage.event_log_storage.get_asset_check_execution_history(
            check_key=check_key,
            limit=limit,
            cursor=cursor,
            status=status,
        )

    def get_latest_asset_check_execution_by_key(
        self,
        check_keys: Sequence["AssetCheckKey"],
    ) -> Mapping["AssetCheckKey", Optional[AssetCheckExecutionRecord]]:
        return self._storage.event_log_storage.get_latest_asset_check_execution_by_key(check_keys)


class LegacyScheduleStorage(ScheduleStorage, ConfigurableClass):
    def __init__(self, storage: DagsterStorage, inst_data: Optional[ConfigurableClassData] = None):
        self._storage = check.inst_param(storage, "storage", DagsterStorage)
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        super().__init__()

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {
            "module_name": str,
            "class_name": str,
            "config_yaml": str,
        }

    @classmethod
    def from_config_value(
        cls, inst_data: Optional[ConfigurableClassData], config_value: Mapping[str, str]
    ) -> "LegacyScheduleStorage":
        storage = ConfigurableClassData(
            module_name=config_value["module_name"],
            class_name=config_value["class_name"],
            config_yaml=config_value["config_yaml"],
        ).rehydrate(as_type=DagsterStorage)
        return LegacyScheduleStorage(storage, inst_data=inst_data)

    @property
    def _instance(self) -> Optional["DagsterInstance"]:
        return self._storage._instance  # noqa: SLF001

    def register_instance(self, instance: "DagsterInstance") -> None:
        if not self._storage.has_instance:
            self._storage.register_instance(instance)

    def wipe(self) -> None:
        return self._storage.schedule_storage.wipe()

    def all_instigator_state(
        self,
        repository_origin_id: Optional[str] = None,
        repository_selector_id: Optional[str] = None,
        instigator_type: Optional["InstigatorType"] = None,
        instigator_statuses: Optional[set["InstigatorStatus"]] = None,
    ) -> Iterable["InstigatorState"]:
        return self._storage.schedule_storage.all_instigator_state(
            repository_origin_id, repository_selector_id, instigator_type, instigator_statuses
        )

    def get_instigator_state(self, origin_id: str, selector_id: str) -> Optional["InstigatorState"]:
        return self._storage.schedule_storage.get_instigator_state(origin_id, selector_id)

    def add_instigator_state(self, state: "InstigatorState") -> "InstigatorState":
        return self._storage.schedule_storage.add_instigator_state(state)

    def update_instigator_state(self, state: "InstigatorState") -> "InstigatorState":
        return self._storage.schedule_storage.update_instigator_state(state)

    def delete_instigator_state(self, origin_id: str, selector_id: str) -> None:
        return self._storage.schedule_storage.delete_instigator_state(origin_id, selector_id)

    @property
    def supports_batch_queries(self) -> bool:
        return self._storage.schedule_storage.supports_batch_queries

    def get_batch_ticks(
        self,
        selector_ids: Sequence[str],
        limit: Optional[int] = None,
        statuses: Optional[Sequence["TickStatus"]] = None,
    ) -> Mapping[str, Iterable["InstigatorTick"]]:
        return self._storage.schedule_storage.get_batch_ticks(selector_ids, limit, statuses)

    def get_tick(self, tick_id: int) -> "InstigatorTick":
        return self._storage.schedule_storage.get_tick(tick_id)

    def get_ticks(
        self,
        origin_id: str,
        selector_id: str,
        before: Optional[float] = None,
        after: Optional[float] = None,
        limit: Optional[int] = None,
        statuses: Optional[Sequence["TickStatus"]] = None,
    ) -> Iterable["InstigatorTick"]:
        return self._storage.schedule_storage.get_ticks(
            origin_id, selector_id, before, after, limit, statuses
        )

    def create_tick(self, tick_data: "TickData") -> "InstigatorTick":
        return self._storage.schedule_storage.create_tick(tick_data)

    def update_tick(self, tick: "InstigatorTick") -> "InstigatorTick":
        return self._storage.schedule_storage.update_tick(tick)

    def purge_ticks(
        self,
        origin_id: str,
        selector_id: str,
        before: float,
        tick_statuses: Optional[Sequence["TickStatus"]] = None,
    ) -> None:
        return self._storage.schedule_storage.purge_ticks(
            origin_id, selector_id, before, tick_statuses
        )

    def add_auto_materialize_asset_evaluations(
        self,
        evaluation_id: int,
        asset_evaluations: Sequence[AutomationConditionEvaluationWithRunIds],
    ) -> None:
        return self._storage.schedule_storage.add_auto_materialize_asset_evaluations(
            evaluation_id, asset_evaluations
        )

    def get_auto_materialize_asset_evaluations(
        self, key: EntityKey, limit: int, cursor: Optional[int] = None
    ) -> Sequence["AutoMaterializeAssetEvaluationRecord"]:
        return self._storage.schedule_storage.get_auto_materialize_asset_evaluations(
            key, limit, cursor
        )

    def get_auto_materialize_evaluations_for_evaluation_id(
        self, evaluation_id: int
    ) -> Sequence["AutoMaterializeAssetEvaluationRecord"]:
        return self._storage.schedule_storage.get_auto_materialize_evaluations_for_evaluation_id(
            evaluation_id
        )

    def purge_asset_evaluations(self, before: float):
        return self._storage.schedule_storage.purge_asset_evaluations(before)

    def upgrade(self) -> None:
        return self._storage.schedule_storage.upgrade()

    def migrate(self, print_fn: Optional[PrintFn] = None, force_rebuild_all: bool = False) -> None:
        return self._storage.schedule_storage.migrate(print_fn, force_rebuild_all)

    def optimize(self, print_fn: Optional[PrintFn] = None, force_rebuild_all: bool = False) -> None:
        return self._storage.schedule_storage.optimize(print_fn, force_rebuild_all)

    def optimize_for_webserver(
        self, statement_timeout: int, pool_recycle: int, max_overflow: int
    ) -> None:
        return self._storage.schedule_storage.optimize_for_webserver(
            statement_timeout,
            pool_recycle,
            max_overflow,
        )

    def dispose(self) -> None:
        return self._storage.schedule_storage.dispose()
